# Copyright © 2016-2019 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract syntax tree for XPath 1.0 expressions with YANG extensions.

This module defines a number of classes that mostly correspond to
variables (non-terminals) of the XPath 1.0 grammar. Only the following
class is intended to be public:

* Expr: XPath 1.0 expression with YANG 1.1 extensions.
"""

import decimal
from math import ceil, copysign, floor
from pyxb.utils.xmlre import XMLToPython, RegularExpressionError
from xml.sax.saxutils import quoteattr
import re
from typing import List, Optional, Tuple
from .schemadata import SchemaContext
from .enumerations import Axis, MultiplicativeOp
from .exceptions import InvalidArgument, XPathTypeError
from .instance import InstanceNode
from .nodeset import NodeExpr, NodeSet, XPathValue
from .typealiases import QualName


class XPathContext:

    def __init__(self, cnode: InstanceNode, origin: InstanceNode,
                 position: int, size: int):
        self.cnode = cnode
        self.origin = origin
        self.position = position
        self.size = size

    def update_cnode(self, new_cnode: InstanceNode) -> "XPathContext":
        return self.__class__(new_cnode, self.origin, self.position, self.size)


class Expr:
    """Abstract class for nodes of XPath AST."""

    indent = 2
    _precedence = 8

    def __str__(self) -> str:
        """Return a string representation of the receiver."""
        raise NotImplementedError

    def evaluate(self, node: InstanceNode) -> XPathValue:
        """Evaluate the receiver and return the result.

        Args:
            node: Context node for XPath evaluation.

        Raises:
            XPathTypeError: If a subexpression of the receiver is of a wrong
                type.
        """
        return self._eval(XPathContext(node, node, 1, 1))

    def _eval_float(self, xctx: XPathContext) -> float:
        val = self._eval(xctx)
        try:
            return float(val)
        except ValueError:
            return float('nan')

    def _eval_string(self, xctx: XPathContext) -> str:
        val = self._eval(xctx)
        if isinstance(val, float):
            try:
                if int(val) == val:
                    return str(int(val))
            except OverflowError:
                return "Infinity" if val > 0 else "-Infinity"
            except ValueError:
                return "NaN"
        if isinstance(val, bool):
            return str(val).lower()
        return str(val)

    def _xfunc_name(self) -> str:
        """Return XPath function name based on the class.

        To be used only for functions.
        """
        fn = self.__class__.__name__[4:]
        return fn[0].lower() + "".join(
            [(c if c.islower() else f"-{c.lower()}") for
             c in fn[1:]])

    def syntax_tree(self, indent: int = 0) -> str:
        """Print abstract syntax tree of the receiver."""
        node_name = self.__class__.__name__
        attr = self._properties_str()
        attr_str = " (" + attr + ")\n" if attr else "\n"
        return (" " * indent + node_name + attr_str +
                self._children_ast(indent + self.indent))

    def _properties_str(self) -> str:
        return ""

    def _children_ast(self, indent) -> str:
        return ""

    def _predicates_str(self, indent) -> str:
        if not self.predicates:
            return ""
        res = " " * indent + "-- Predicates:\n"
        newi = indent + 3
        for p in self.predicates:
            res += p.syntax_tree(newi)
        return res

    def _apply_predicates(self, ns: XPathValue,
                          xctx: XPathContext) -> XPathValue:
        for p in self.predicates:
            res = NodeSet([])
            size = len(ns)
            for i in range(size):
                pval = p._eval(XPathContext(ns[i], xctx.origin, i + 1, size))
                try:
                    if isinstance(pval, float) and pval > 0:
                        res.append(ns[int(pval) - 1])
                        break
                except IndexError:
                    return res
                if pval:
                    res.append(ns[i])
            ns = res
        return ns


class UnaryExpr(Expr):
    """Abstract superclass for unary expressions."""

    def __init__(self, expr: Optional[Expr]):
        self.expr = expr

    def __str__(self) -> str:
        """Return string representation of a unary function.

        Non-function subclasses should override this method.
        """
        return f"{self._xfunc_name()}({self.expr if self.expr else ''})"

    def _children_ast(self, indent: int) -> str:
        return self.expr.syntax_tree(indent) if self.expr else ""


class BinaryExpr(Expr):
    """Abstract superclass of binary expressions."""

    def __init__(self, left: Expr, right: Expr):
        self.left = left
        self.right = right

    def __str__(self) -> str:
        """Return string representation of a binary function.

        Non-function subclasses should override this method.
        """
        return f"{self._xfunc_name()}({self.left}, {self.right})"

    def _children_ast(self, indent: int) -> str:
        return self.left.syntax_tree(indent) + self.right.syntax_tree(indent)

    def _eval_ops(self, xctx: XPathContext) -> Tuple[XPathValue, XPathValue]:
        return (self.left._eval(xctx), self.right._eval(xctx))

    def _eval_ops_float(self, xctx: XPathContext) -> Tuple[float, float]:
        return (self.left._eval_float(xctx), self.right._eval_float(xctx))

    def _eval_ops_string(self, xctx: XPathContext) -> Tuple[str, str]:
        return (self.left._eval_string(xctx), self.right._eval_string(xctx))

    def _as_str(self, op: str, spaces=True) -> str:
        lft = str(self.left)
        if self.left._precedence < self._precedence:
            lft = "(" + lft + ")"
        rt = str(self.right)
        if self.right._precedence < self._precedence:
            rt = "(" + rt + ")"
        sop = f" {op} " if spaces else op
        return f"{lft}{sop}{rt}"


class OrExpr(BinaryExpr):

    _precedence = 0

    def __str__(self) -> str:
        return self._as_str("or")

    def _eval(self, xctx: XPathContext) -> bool:
        lres, rres = self._eval_ops(xctx)
        return lres or rres


class AndExpr(BinaryExpr):

    _precedence = 1

    def __str__(self) -> str:
        return self._as_str("and")

    def _eval(self, xctx: XPathContext) -> bool:
        lres, rres = self._eval_ops(xctx)
        return lres and rres


class EqualityExpr(BinaryExpr):

    _precedence = 2

    def __init__(self, left: Expr, right: Expr, negate: bool):
        super().__init__(left, right)
        self.negate = negate

    def __str__(self) -> str:
        return self._as_str("!=" if self.negate else "=")

    def _properties_str(self) -> str:
        return "!=" if self.negate else "="

    def _eval(self, xctx: XPathContext) -> bool:
        lres, rres = self._eval_ops(xctx)
        return lres != rres if self.negate else lres == rres


class RelationalExpr(BinaryExpr):

    _precedence = 3

    def __init__(self, left: Expr, right: Expr, less: bool,
                 equal: bool):
        super().__init__(left, right)
        self.less = less
        self.equal = equal

    def __str__(self) -> str:
        lg = "<" if self.less else ">"
        return self._as_str(lg + "=" if self.equal else lg)

    def _properties_str(self) -> str:
        res = "<" if self.less else ">"
        if self.equal:
            res += "="
        return res

    def _eval(self, xctx: XPathContext) -> bool:
        lres, rres = self._eval_ops(xctx)
        if self.less:
            return lres <= rres if self.equal else lres < rres
        return lres >= rres if self.equal else lres > rres


class AdditiveExpr(BinaryExpr):

    _precedence = 4

    def __init__(self, left: Expr, right: Expr, plus: bool):
        super().__init__(left, right)
        self.plus = plus

    def __str__(self) -> str:
        return self._as_str("+" if self.plus else "-")

    def _properties_str(self) -> str:
        return "+" if self.plus else "-"

    def _eval(self, xctx: XPathContext) -> float:
        lres, rres = self._eval_ops_float(xctx)
        return lres + rres if self.plus else lres - rres


class MultiplicativeExpr(BinaryExpr):

    _precedence = 5

    def __init__(self, left: Expr, right: Expr,
                 operator: MultiplicativeOp):
        super().__init__(left, right)
        self.operator = operator

    def __str__(self) -> str:
        return self._as_str(str(self.operator))

    def _properties_str(self) -> str:
        if self.operator == MultiplicativeOp.multiply:
            return "*"
        if self.operator == MultiplicativeOp.divide:
            return "div"
        if self.operator == MultiplicativeOp.modulo:
            return "mod"

    def _eval(self, xctx: XPathContext) -> float:
        lres, rres = self._eval_ops_float(xctx)
        if self.operator == MultiplicativeOp.multiply:
            return lres * rres
        if self.operator == MultiplicativeOp.divide:
            try:
                return lres / rres
            except ZeroDivisionError:
                return (float("nan") if lres == 0.0
                        else copysign(float('inf'), lres))
        try:
            return copysign(lres % rres, lres)
        except ZeroDivisionError:
            return float('nan')


class UnaryMinusExpr(UnaryExpr):

    _precedence = 6

    def __init__(self, expr: Expr, negate: bool):
        super().__init__(expr)
        self.negate = negate

    def __str__(self) -> str:
        return f"{'-' if self.negate else ''}{self.expr}"

    def _properties_str(self) -> str:
        return "-" if self.negate else "+"

    def _eval(self, xctx: XPathContext) -> float:
        res = self.expr._eval_float(xctx)
        return -res if self.negate else res


class UnionExpr(BinaryExpr):

    _precedence = 7

    def __str__(self) -> str:
        return self._as_str("|")

    def _eval(self, xctx: XPathContext) -> NodeSet:
        lres, rres = self._eval_ops(xctx)
        return lres.union(rres)


class Literal(Expr):

    def __init__(self, value: str):
        self.value = value

    def __str__(self) -> str:
        return quoteattr(self.value)

    def _properties_str(self) -> str:
        return self.value

    def _eval(self, xctx: XPathContext) -> str:
        return self.value


class Number(Expr):

    def __init__(self, value: float):
        self.value = value

    def __str__(self) -> str:
        return str(self.value)

    def _properties_str(self) -> str:
        return str(self.value)

    def _eval(self, xctx: XPathContext) -> float:
        return float(self.value)


class PathExpr(BinaryExpr):

    def __str__(self) -> str:
        return self._as_str("/", spaces=False)

    def _eval(self, xctx: XPathContext) -> XPathValue:
        ns = self.left._eval(xctx)
        if not isinstance(ns, NodeSet):
            raise XPathTypeError(str(ns))
        res = NodeSet([])
        for n in ns:
            res = res.union(self.right._eval(xctx.update_cnode(n)))
        return res


class FilterExpr(Expr):

    def __init__(self, primary: Expr, predicates: List[Expr]):
        self.primary = primary
        self.predicates = predicates

    def __str__(self) -> str:
        return (str(self.primary) +
                "".join([f"[{p}]" for p in self.predicates]))

    def _children_ast(self, indent) -> str:
        return self.primary.syntax_tree(indent) + self._predicates_str(indent)

    def _eval(self, xctx: XPathContext) -> XPathValue:
        res = self.primary._eval(xctx)
        return self._apply_predicates(res, xctx)


class LocationPath(BinaryExpr):

    def _eval(self, xctx: XPathContext) -> XPathValue:
        lres = self.left._eval(xctx)
        ns = lres.bind(self.right._node_trans(xctx))
        return self.right._apply_predicates(ns, xctx)

    def __str__(self) -> str:
        sep = "" if isinstance(self.left, Root) else "/"
        return f"{self.left}{sep}{self.right}"


class Root(Expr):

    _precedence = 8

    def _eval(self, xctx: XPathContext) -> NodeSet:
        return NodeSet([xctx.cnode.top()])

    def __str__(self) -> str:
        return "/"


class Step(Expr):

    def __init__(self, axis: Axis, qname: QualName,
                 predicates: List[Expr]):
        self.axis = axis
        self.qname = qname
        self.predicates = predicates

    def __str__(self) -> str:
        if self.axis == Axis.descendant_or_self and self.qname is None:
            return ""
        if self.axis == Axis.self and self.qname is None:
            return "."
        if self.axis == Axis.parent and self.qname is None:
            return ".."
        if self.qname:
            qn = (f"{self.qname[1]}:{self.qname[0]}" if self.qname[1]
                  else self.qname[0])
        elif self.qname is None:
            qn = "node()"
        else:
            qn = "*"
        ax = f"{self.axis}::" if self.axis != Axis.child else ""
        prs = "".join([f"[{p}]" for p in self.predicates])
        return ax + qn + prs

    def _properties_str(self) -> str:
        return f"{self.axis.name} {self.qname}"

    def _children_ast(self, indent) -> str:
        return self._predicates_str(indent)

    def _node_trans(self, xctx) -> NodeExpr:
        qname = ((self.qname[0], xctx.origin.namespace) if
                 self.qname and self.qname[1] is None else self.qname)
        return {
            Axis.ancestor: lambda n, qn=qname: n._ancestors(qn),
            Axis.ancestor_or_self:
                lambda n, qn=qname: n._ancestors_or_self(qn),
            Axis.child: lambda n, qn=qname: n._children(qn),
            Axis.descendant: lambda n, qn=qname: n._descendants(qn),
            Axis.descendant_or_self:
                lambda n, qn=qname: n._descendants(qn, True),
            Axis.following_sibling:
                lambda n, qn=qname: n._following_siblings(qn),
            Axis.parent: (
                lambda n, qn=qname: [] if qn and qn != n.parent.qual_name
                else n._parent()),
            Axis.preceding_sibling:
                lambda n, qn=qname: n._preceding_siblings(qn),
            Axis.self:
                lambda n, qn=qname: [] if qn and qn != n.qual_name else [n],
        }[self.axis]

    def _eval(self, xctx: XPathContext) -> XPathValue:
        ns = NodeSet(self._node_trans(xctx)(xctx.cnode))
        return self._apply_predicates(ns, xctx)


class FuncBitIsSet(BinaryExpr):

    def _eval(self, xctx: XPathContext) -> bool:
        ns = self.left._eval(xctx)
        if not isinstance(ns, NodeSet):
            raise XPathTypeError(str(ns))
        bit = self.right._eval_string(xctx)
        try:
            return bit in ns[0].value
        except (IndexError, TypeError):
            return False


class FuncBoolean(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> bool:
        return bool(self.expr._eval(xctx))


class FuncCeiling(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> float:
        return float(ceil(self.expr._eval_float(xctx)))


class FuncConcat(Expr):

    def __init__(self, parts: List[Expr]):
        self.parts = parts

    def __str__(self) -> str:
        return f"concat({', '.join([str(p) for p in self.parts])})"

    def _children_ast(self, indent: int) -> str:
        return "".join([ex.syntax_tree(indent) for ex in self.parts])

    def _eval(self, xctx: XPathContext) -> str:
        return "".join([ex._eval_string(xctx) for ex in self.parts])


class FuncContains(BinaryExpr):

    def _eval(self, xctx: XPathContext) -> bool:
        lres, rres = self._eval_ops_string(xctx)
        return lres.find(rres) >= 0


class FuncCount(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> int:
        ns = self.expr._eval(xctx)
        return float(len(ns))


class FuncCurrent(Expr):

    def _eval(self, xctx: XPathContext) -> NodeSet:
        return NodeSet([xctx.origin])

    def __str__(self) -> str:
        return "current()"


class FuncDeref(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> NodeSet:
        ns = self.expr._eval(xctx)
        if not isinstance(ns, NodeSet):
            raise XPathTypeError(str(ns))
        ref = ns[0]
        return NodeSet(ref._deref())


class FuncDerivedFrom(BinaryExpr):

    def __init__(self, left: Expr, right: Expr, or_self: bool,
                 sctx: SchemaContext):
        super().__init__(left, right)
        self.or_self = or_self
        self.sctx = sctx

    def _xfunc_name(self) -> str:
        return ("derived-from-or-self" if self.or_self else "derived-from")

    def _properties_str(self) -> str:
        return ("OR-SELF, " if self.or_self
                else "") + self.sctx.schema_data.namespace(self.mid)

    def _eval(self, xctx: XPathContext) -> bool:
        ns = self.left._eval(xctx)
        if not isinstance(ns, NodeSet):
            raise XPathTypeError(str(ns))
        i = self.sctx.schema_data.translate_pname(
            self.right._eval_string(xctx), self.sctx.text_mid)
        for n in ns:
            if not n.schema_node._is_identityref():
                return False
            if self.or_self and n.value == i:
                return True
            if self.sctx.schema_data.is_derived_from(n.value, i):
                return True
        return False


class FuncEnumValue(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> float:
        ns = self.expr._eval(xctx)
        if not isinstance(ns, NodeSet):
            raise XPathTypeError(str(ns))
        try:
            node = ns[0]
            return float(node.schema_node.type.enum[node.value])
        except (AttributeError, IndexError, KeyError):
            return float('nan')


class FuncFalse(Expr):

    def __str__(self) -> str:
        return "false()"

    def _eval(self, xctx: XPathContext) -> bool:
        return False


class FuncFloor(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> float:
        return float(floor(self.expr._eval_float(xctx)))


class FuncLast(Expr):

    def __str__(self) -> str:
        return "last()"

    def _eval(self, xctx: XPathContext) -> int:
        return float(xctx.size)


class FuncName(UnaryExpr):

    def __init__(self, expr: Optional[Expr], local: bool):
        super().__init__(expr)
        self.local = local

    def __str__(self) -> str:
        fn = "local-name" if self.local else "name"
        return f"{fn}({self.expr if self.expr else ''})"

    def _properties_str(self) -> str:
        return "LOCAL" if self.local else ""

    def _eval(self, xctx: XPathContext) -> str:
        if self.expr is None:
            node = xctx.cnode
        else:
            ns = self.expr._eval(xctx)
            try:
                node = ns[0]
            except TypeError:
                raise XPathTypeError(str(ns))
            except IndexError:
                return ""
        if node.path == ():
            return ""
        if self.local:
            p, s, loc = node.name.partition(":")
            return loc if s else p
        return node.name


class FuncNormalizeSpace(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> str:
        string = self.expr._eval_string(xctx) if self.expr else str(xctx.cnode)
        return " ".join(string.strip().split())


class FuncNot(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> bool:
        return not(self.expr._eval(xctx))


class FuncNumber(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> float:
        if self.expr is None:
            try:
                return float(xctx.cnode.value)
            except ValueError:
                return float('nan')
        return self.expr._eval_float(xctx)


class FuncPosition(Expr):

    def __str__(self) -> str:
        return "position()"

    def _eval(self, xctx: XPathContext) -> int:
        return xctx.position


class FuncReMatch(BinaryExpr):

    def _eval(self, xctx: XPathContext) -> bool:
        lres, rres = self._eval_ops_string(xctx)
        try:
            return re.match(XMLToPython(rres), lres) is not None
        except RegularExpressionError:
            raise InvalidArgument(rres) from None


class FuncRound(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> float:
        dec = decimal.Decimal(self.expr._eval_float(xctx))
        try:
            return float(dec.to_integral_value(
                decimal.ROUND_HALF_UP if dec > 0 else decimal.ROUND_HALF_DOWN))
        except decimal.InvalidOperation:
            return float('nan')


class FuncStartsWith(BinaryExpr):

    def _eval(self, xctx: XPathContext) -> bool:
        lres, rres = self._eval_ops_string(xctx)
        return lres.startswith(rres)


class FuncString(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> str:
        if self.expr is None:
            return str(xctx.cnode)
        return self.expr._eval_string(xctx)


class FuncStringLength(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> str:
        string = self.expr._eval_string(xctx) if self.expr else str(xctx.cnode)
        return float(len(string))


class FuncSubstring(BinaryExpr):

    def __init__(self, string: Expr, start: Expr,
                 length: Optional[Expr]):
        super().__init__(string, start)
        self.length = length

    def __str__(self) -> str:
        res = f"{self._xfunc_name()}({self.left}, {self.right}"
        if self.length:
            res += f", {self.length}"
        return res + ")"

    def _children_ast(self, indent: int) -> str:
        return super()._children_ast(indent) + self.length.syntax_tree(indent)

    def _eval(self, xctx: XPathContext) -> str:
        string = self.left._eval_string(xctx)
        rres = self.right._eval_float(xctx)
        try:
            start = round(rres) - 1
        except (ValueError, OverflowError):
            return "" if self.length or rres != float("-inf") else string
        if self.length is None:
            return string[max(start, 0):]
        length = self.length._eval_float(xctx)
        try:
            end = start + round(length)
        except (ValueError, OverflowError):
            return string[max(start, 0):] if length == float('inf') else ""
        return string[max(start, 0):end]


class FuncSubstringAfter(BinaryExpr):

    def _eval(self, xctx: XPathContext) -> str:
        lres, rres = self._eval_ops_string(xctx)
        ind = lres.find(rres)
        return lres[ind + len(rres):] if ind >= 0 else ""


class FuncSubstringBefore(BinaryExpr):

    def _eval(self, xctx: XPathContext) -> str:
        lres, rres = self._eval_ops_string(xctx)
        ind = lres.find(rres)
        return lres[:ind] if ind >= 0 else ""


class FuncSum(UnaryExpr):

    def _eval(self, xctx: XPathContext) -> float:
        ns = self.expr._eval(xctx)
        if not isinstance(ns, NodeSet):
            raise XPathTypeError(str(ns))
        try:
            return float(sum([n.value for n in ns]))
        except TypeError:
            return float('nan')


class FuncTranslate(BinaryExpr):

    def __init__(self, s1: Expr, s2: Expr, s3: Expr):
        super().__init__(s1, s2)
        self.nchars = s3

    def __str__(self) -> str:
        return f"{self._xfunc_name()}({self.left}, {self.right}, {self.nchars})"

    def _children_ast(self, indent: int) -> str:
        return super()._children_ast(indent) + self.nchars.syntax_tree(indent)

    def _eval(self, xctx: XPathContext) -> str:
        string, old = self._eval_ops_string(xctx)
        new = self.nchars._eval_string(xctx)[:len(old)]
        ttab = str.maketrans(old[:len(new)], new, old[len(new):])
        return string.translate(ttab)


class FuncTrue(Expr):

    def __str__(self) -> str:
        return "true()"

    def _eval(self, xctx: XPathContext) -> bool:
        return True



# Copyright © 2016-2019 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""YANG schema nodes.

This module implements the following classes:

* SchemaNode: Abstract class for schema nodes.
* InternalNode: Abstract class for schema nodes that have children.
* GroupNode: Anonymous group of schema nodes.
* SchemaTreeNode: Root node of a schema tree.
* DataNode: Abstract class for data nodes.
* TerminalNode: Abstract class for schema nodes that have no children.
* ContainerNode: YANG container node.
* SequenceNode: Abstract class for schema nodes that represent a sequence.
* ListNode: YANG list node.
* ChoiceNode: YANG choice node.
* CaseNode: YANG case node.
* RpcActionNode: YANG rpc or action node.
* InputNode: YANG input node.
* OutputNode: YANG output node.
* NotificationNode: YANG notification node.
* LeafNode: YANG leaf node.
* LeafListNode: YANG leaf-list node.
* AnyContentNode: Abstract superclass for anydata and anyxml nodes..
* AnydataNode: YANG anydata node.
* AnyxmlNode: YANG anyxml node.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Set, Tuple
from .constraint import Must
from .datatype import (DataType, LinkType,
                       RawScalar, IdentityrefType)
from .enumerations import Axis, ContentType, DefaultDeny, ValidationScope
from .exceptions import (
    AnnotationTypeError, InvalidArgument,
    MissingAnnotationTarget, MissingAugmentTarget, RawMemberError,
    RawTypeError, SchemaError, SemanticError, UndefinedAnnotation,
    YangsonException, YangTypeError)
from .instvalue import (
    ArrayValue, EntryValue, MetadataObject, ObjectValue, Value)
from .schemadata import IdentityAdjacency, SchemaContext
from .schpattern import (ChoicePattern, ConditionalPattern, Empty, Member,
                         NotAllowed, Pair, SchemaPattern)
from .statement import Statement
from .typealiases import (DataPath, InstanceName, JSONPointer, QualName,
                          RawEntry, RawList, RawObject, RawValue,
                          RawMetadataObject, ScalarValue, SchemaRoute,
                          YangIdentifier)
from .xpathparser import XPathParser


class Annotation:
    """Class for metadata annotations [RFC 7952]."""

    def __init__(self, type: "DataType", description: str = None):
        """Initialize the class instance."""
        self.type = type
        self.description = description


class SchemaNode:
    """Abstract class for all schema nodes."""

    def __init__(self):
        """Initialize the class instance."""
        self.name = None  # type: Optional[YangIdentifier]
        """Name of the receiver."""
        self.ns = None  # type: Optional[YangIdentifier]
        """Namespace of the receiver."""
        self.parent = None  # type: Optional["InternalNode"]
        """Parent schema node."""
        self.description = None  # type: Optional[str]
        """Description of the receiver."""
        self.must = []  # type: List[Must]
        """List of "must" expressions attached to the receiver."""
        self.when = None  # type: Optional["Expr"]
        """Optional "when" expression that makes the receiver conditional."""
        self.val_count = 0
        self._ctype = None
        """Content type of the receiver."""

    @property
    def qual_name(self) -> QualName:
        """Qualified name of the receiver."""
        return (self.name, self.ns)

    @property
    def config(self) -> bool:
        """Does the receiver (also) represent configuration?"""
        return self.content_type().value & ContentType.config.value != 0

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return False

    def schema_root(self) -> "SchemaTreeNode":
        """Return the root node of the receiver's schema."""
        sn = self
        while sn.parent:
            sn = sn.parent
        return sn

    def content_type(self) -> ContentType:
        """Return receiver's content type."""
        return self._ctype if self._ctype else self.parent.content_type()

    def data_parent(self) -> Optional["InternalNode"]:
        """Return the closest ancestor data node."""
        parent = self.parent
        while parent:
            if isinstance(parent, DataNode):
                return parent
            parent = parent.parent

    def iname(self) -> InstanceName:
        """Return the instance name corresponding to the receiver."""
        dp = self.data_parent()
        return (self.name if dp and self.ns == dp.ns
                else self.ns + ":" + self.name)

    def data_path(self) -> DataPath:
        """Return the receiver's data path."""
        dp = self.data_parent()
        return (dp.data_path() if dp else "") + "/" + self.iname()

    def state_roots(self) -> List[DataPath]:
        """Return a list of data paths to descendant state data roots."""
        return [r.data_path() for r in self._state_roots()]

    def from_raw(self, rval: RawValue, jptr: JSONPointer = "") -> Value:
        """Return instance value transformed from a raw value using receiver.

        Args:
            rval: Raw value.
            jptr: JSON pointer of the current instance node.

        Raises:
            RawMemberError: If a member inside `rval` is not defined in the
                schema.
            RawTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        raise NotImplementedError

    def clear_val_counters(self) -> None:
        """Clear receiver's validation counter."""
        self.val_count = 0

    def _get_description(self, stmt: Statement):
        dst = stmt.find1("description")
        if dst is not None:
            self.description = dst.argument

    def _yang_class(self) -> str:
        return self.__class__.__name__[:-4].lower()

    def _node_digest(self) -> Dict[str, Any]:
        """Return dictionary of receiver's properties suitable for clients."""
        res = {"kind": self._yang_class()}
        if self.mandatory:
            res["mandatory"] = True
        if self.description:
            res["description"] = self.description
        return res

    def _validate(self, inst: "InstanceNode", scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Validate instance against the receiver.

        Args:
            inst: Instance node to be validated.
            scope: Scope of the validation (syntax, semantics or all)
            ctype: Content type of the instance.

        Returns:
            ``None`` if validation succeeds.

        Raises:
            SchemaError: if `inst` doesn't conform to the schema.
            SemanticError: If `inst` violates a semantic rule.
            YangTypeError: If `inst` is a scalar of incorrect type.
        """
        self.val_count += 1

    def _iname2qname(self, iname: InstanceName) -> QualName:
        """Translate instance name to qualified name in the receiver's context.
        """
        p, s, loc = iname.partition(":")
        return (loc, p) if s else (p, self.ns)

    def _flatten(self) -> List["SchemaNode"]:
        return [self]

    def _handle_substatements(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Dispatch actions for substatements of `stmt`."""
        for s in stmt.substatements:
            if s.prefix:
                key = (
                    sctx.schema_data.modules[sctx.text_mid].prefix_map[s.prefix][0]
                    + ":" + s.keyword)
            else:
                key = s.keyword
            mname = SchemaNode._stmt_callback.get(key, "_noop")
            method = getattr(self, mname)
            method(s, sctx)

    def _follow_leafref(
            self, xpath: "Expr", init: "TerminalNode") -> Optional["DataNode"]:
        """Return the data node referred to by a leafref path.

        Args:
            xpath: XPath expression compiled from a leafref path.
            init: initial context node
        """
        if isinstance(xpath, LocationPath):
            lft = self._follow_leafref(xpath.left, init)
            if lft is None:
                return None
            return lft._follow_leafref(xpath.right, init)
        elif isinstance(xpath, Step):
            if xpath.axis == Axis.parent:
                if isinstance(self, SchemaTreeNode):
                    return None
                return self.data_parent() or self.schema_root()
            elif xpath.axis == Axis.child:
                if isinstance(self, InternalNode) and xpath.qname:
                    qname = (xpath.qname if xpath.qname[1]
                             else (xpath.qname[0], init.ns))
                    return self.get_data_child(*qname)
        elif isinstance(xpath, Root):
            return self.schema_root()
        return None

    def _noop(self, stmt: Statement, sctx: SchemaContext) -> None:
        pass

    def _config_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        if stmt.argument == "false":
            self._ctype = ContentType.nonconfig

    def _description_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.description = stmt.argument

    def _must_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        xpp = XPathParser(stmt.argument, sctx)
        mex = xpp.parse()
        if not xpp.at_end():
            raise InvalidArgument(stmt.argument)
        self.must.append(Must(mex, *stmt.get_error_info()))

    def _when_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        xpp = XPathParser(stmt.argument, sctx)
        wex = xpp.parse()
        if not xpp.at_end():
            raise InvalidArgument(stmt.argument)
        self.when = wex

    def _mandatory_stmt(self, stmt, sctx: SchemaContext) -> None:
        if stmt.argument == "true":
            self._mandatory = True
        elif stmt.argument == "false":
            self._mandatory = False

    def _post_process(self) -> None:
        pass

    def _is_identityref(self) -> bool:
        return False

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        return []

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return self._tree_line_prefix() + " " + self.iname()

    def _tree_line_prefix(self) -> str:
        return "+--"

    def _nacm_default_deny_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Set NACM default access."""
        if not hasattr(self, 'default_deny'):
            return
        if stmt.keyword == "default-deny-all":
            self.default_deny = DefaultDeny.all
        elif stmt.keyword == "default-deny-write":
            self.default_deny = DefaultDeny.write

    _stmt_callback = {
        "action": "_rpc_action_stmt",
        "anydata": "_anydata_stmt",
        "anyxml": "_anydata_stmt",
        "case": "_case_stmt",
        "choice": "_choice_stmt",
        "config": "_config_stmt",
        "container": "_container_stmt",
        "default": "_default_stmt",
        "identity": "_identity_stmt",
        "ietf-netconf-acm:default-deny-all": "_nacm_default_deny_stmt",
        "ietf-netconf-acm:default-deny-write": "_nacm_default_deny_stmt",
        "ietf-yang-metadata:annotation": "_annotation_stmt",
        "input": "_input_stmt",
        "key": "_key_stmt",
        "leaf": "_leaf_stmt",
        "leaf-list": "_leaf_list_stmt",
        "list": "_list_stmt",
        "mandatory": "_mandatory_stmt",
        "max-elements": "_max_elements_stmt",
        "min-elements": "_min_elements_stmt",
        "must": "_must_stmt",
        "notification": "_notification_stmt",
        "output": "_output_stmt",
        "ordered-by": "_ordered_by_stmt",
        "presence": "_presence_stmt",
        "rpc": "_rpc_action_stmt",
        "unique": "_unique_stmt",
        "uses": "_uses_stmt",
        "when": "_when_stmt",
    }
    """Map of statement keywords to callback methods."""


class InternalNode(SchemaNode):
    """Abstract class for schema nodes that have children."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.children = []  # type: List[SchemaNode]
        self._mandatory_children = set()  # type: MutableSet[SchemaNode]

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return len(self._mandatory_children) > 0

    def get_child(self, name: YangIdentifier,
                  ns: YangIdentifier = None) -> Optional[SchemaNode]:
        """Return receiver's schema child.

        Args:
            name: Child's name.
            ns: Child's namespace (= `self.ns` if absent).
        """
        ns = ns if ns else self.ns
        todo = []
        for child in self.children:
            if child.name is None:
                todo.append(child)
            elif child.name == name and child.ns == ns:
                return child
        for c in todo:
            grandchild = c.get_child(name, ns)
            if grandchild is not None:
                return grandchild

    def get_schema_descendant(
            self, route: SchemaRoute) -> Optional[SchemaNode]:
        """Return descendant schema node or ``None`` if not found.

        Args:
            route: Schema route to the descendant node
                   (relative to the receiver).
        """
        node = self
        for p in route:
            try:
                node = node.get_child(*p)
            except AttributeError:
                return None
            if node is None:
                return None
        return node

    def get_data_child(self, name: YangIdentifier,
                       ns: YangIdentifier = None) -> Optional["DataNode"]:
        """Return data node directly under the receiver."""
        ns = ns if ns else self.ns
        todo = []
        for child in self.children:
            if child.name == name and child.ns == ns:
                if isinstance(child, DataNode):
                    return child
                todo.insert(0, child)
            elif not isinstance(child, DataNode):
                todo.append(child)
        for c in todo:
            res = c.get_data_child(name, ns)
            if res:
                return res

    def filter_children(self, ctype: ContentType = None) -> List[SchemaNode]:
        """Return receiver's children based on content type.

        Args:
            ctype: Content type.
        """
        if ctype is None:
            ctype = self.content_type()
        return [c for c in self.children if
                not isinstance(c, (RpcActionNode, NotificationNode)) and
                c.content_type().value & ctype.value != 0]

    def data_children(self) -> List["DataNode"]:
        """Return the set of all data nodes directly under the receiver."""
        res = []
        for child in self.children:
            if isinstance(child, DataNode):
                res.append(child)
            elif not isinstance(child, SchemaTreeNode):
                res.extend(child.data_children())
        return res

    def from_raw(self, rval: RawObject, jptr: JSONPointer = "") -> ObjectValue:
        """Override the superclass method."""
        if not isinstance(rval, dict):
            raise RawTypeError(jptr, "object")
        res = ObjectValue()
        for qn in rval:
            if qn.startswith("@"):
                if qn != "@":
                    tgt = qn[1:]
                    if tgt not in rval:
                        raise MissingAnnotationTarget(jptr, tgt)
                    jptr += '/' + tgt
                res[qn] = self._process_metadata(rval[qn], jptr)
            else:
                cn = self._iname2qname(qn)
                ch = self.get_data_child(*cn)
                npath = jptr + "/" + qn
                if ch is None:
                    raise RawMemberError(npath)
                res[ch.iname()] = ch.from_raw(rval[qn], npath)
        return res

    def _process_metadata(self, rmo: RawMetadataObject, jptr: JSONPointer) -> MetadataObject:
        res = {}
        ans = self.schema_root().annotations
        for mem in rmo:
            try:
                an = ans[self._iname2qname(mem)]
            except KeyError:
                raise UndefinedAnnotation(jptr, mem)
            res[mem] = an.type.from_raw(rmo[mem])
            if res[mem] not in an.type:
                raise AnnotationTypeError(jptr, mem, an.type.error_message)
        return res

    def _node_digest(self) -> Dict[str, Any]:
        res = super()._node_digest()
        rc = res["children"] = {}
        for c in self.data_children():
            cdig = rc[c.iname()] = c._node_digest()
            if self.config and not c.config:
                cdig["config"] = False
        for c in [c for c in self.children if isinstance(c, SchemaTreeNode)]:
            rc[c.iname()] = c._node_digest()
        return res

    def _validate(self, inst: "InstanceNode", scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Extend the superclass method."""
        if scope.value & ValidationScope.syntax.value:   # schema
            self._check_schema_pattern(inst, ctype)
        for m in inst:
            inst._member(m).validate(scope, ctype)
        super()._validate(inst, scope, ctype)

    def _add_child(self, node: SchemaNode) -> None:
        node.parent = self
        self.children.append(node)

    def _child_inst_names(self) -> Set[InstanceName]:
        """Return the set of instance names under the receiver."""
        return frozenset([c.iname() for c in self.data_children()])

    def _check_schema_pattern(self, inst: "InstanceNode",
                              ctype: ContentType) -> None:
        p = self.schema_pattern
        p._eval_when(inst)
        for m in inst:
            newp = p.deriv(m, ctype)
            if isinstance(newp, NotAllowed):
                raise SchemaError(
                    inst.json_pointer(),
                    ("" if ctype == ContentType.all else ctype.name + " ") +
                    "member-not-allowed", m)
            p = newp
        if not p.nullable(ctype):
            mms = p._mandatory_members(ctype)
            msg = "one of " if len(mms) > 1 else ""
            raise SchemaError(inst.json_pointer(), "missing-data",
                              "expected " + msg + ", ".join([repr(m) for m in mms]))

    def _make_schema_patterns(self) -> None:
        """Build schema pattern for the receiver and its data descendants."""
        self.schema_pattern = self._schema_pattern()
        for dc in self.data_children():
            if isinstance(dc, InternalNode):
                dc._make_schema_patterns()

    def _schema_pattern(self) -> SchemaPattern:
        todo = [c for c in self.children
                if not isinstance(c, (RpcActionNode, NotificationNode))]
        if not todo:
            return Empty()
        prev = todo[0]._pattern_entry()
        for c in todo[1:]:
            prev = Pair(c._pattern_entry(), prev)
        return ConditionalPattern(prev, self.when) if self.when else prev

    def _post_process(self) -> None:
        super()._post_process()
        for c in self.children:
            c._post_process()

    def _add_mandatory_child(self, node: SchemaNode) -> None:
        """Add `node` to the set of mandatory children."""
        self._mandatory_children.add(node)

    def _add_defaults(self, inst: "InstanceNode", ctype: ContentType,
                      lazy: bool = False) -> "InstanceNode":
        for c in self.filter_children(ctype):
            if isinstance(c, DataNode):
                inst = c._default_instance(inst, ctype, lazy)
            elif not isinstance(c, (RpcActionNode, NotificationNode)):
                inst = c._add_defaults(inst, ctype)
        return inst

    def _state_roots(self) -> List[SchemaNode]:
        if self.content_type() == ContentType.nonconfig:
            return [self]
        res = []
        for c in self.data_children():
            res.extend(c._state_roots())
        return res

    def _handle_child(
            self, node: SchemaNode, stmt: Statement, sctx: SchemaContext) -> None:
        """Add child node to the receiver and handle substatements."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        node.name = stmt.argument
        node.ns = sctx.default_ns
        node._get_description(stmt)
        self._add_child(node)
        node._handle_substatements(stmt, sctx)

    def _augment_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle **augment** statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        path = sctx.schema_data.sni2route(stmt.argument, sctx)
        target = self.get_schema_descendant(path)
        if target is None:
            raise MissingAugmentTarget(stmt.argument)
        if stmt.find1("when"):
            gr = GroupNode()
            target._add_child(gr)
            target = gr
        target._handle_substatements(stmt, sctx)

    def _refine_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle **refine** statement."""
        target = self.get_schema_descendant(
            sctx.schema_data.sni2route(stmt.argument, sctx))
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            target.parent.children.remove(target)
        else:
            target._handle_substatements(stmt, sctx)

    def _uses_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle uses statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        grp, gid = sctx.schema_data.get_definition(stmt, sctx)
        wst = stmt.find1("when")
        if wst:
            sn = GroupNode()
            xpp = XPathParser(wst.argument, sctx)
            wex = xpp.parse()
            if not xpp.at_end():
                raise InvalidArgument(wst.argument)
            sn.when = wex
            self._add_child(sn)
        else:
            sn = self
        sn._handle_substatements(grp, gid)
        for augst in stmt.find_all("augment"):
            sn._augment_stmt(augst, sctx)
        for refst in stmt.find_all("refine"):
            sn._refine_stmt(refst, sctx)

    def _container_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle container statement."""
        self._handle_child(ContainerNode(), stmt, sctx)

    def _identity_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle identity statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        id = (stmt.argument, sctx.schema_data.namespace(sctx.text_mid))
        adj = sctx.schema_data.identity_adjs.setdefault(id, IdentityAdjacency())
        for bst in stmt.find_all("base"):
            bid = sctx.schema_data.translate_pname(bst.argument, sctx.text_mid)
            adj.bases.add(bid)
            badj = sctx.schema_data.identity_adjs.setdefault(
                bid, IdentityAdjacency())
            badj.derivs.add(id)
        sctx.schema_data.identity_adjs[id] = adj

    def _list_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle list statement."""
        self._handle_child(ListNode(), stmt, sctx)

    def _choice_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle choice statement."""
        self._handle_child(ChoiceNode(), stmt, sctx)

    def _case_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle case statement."""
        self._handle_child(CaseNode(), stmt, sctx)

    def _leaf_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle leaf statement."""
        node = LeafNode()
        node.type = DataType._resolve_type(
            stmt.find1("type", required=True), sctx)
        self._handle_child(node, stmt, sctx)

    def _leaf_list_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle leaf-list statement."""
        node = LeafListNode()
        node.type = DataType._resolve_type(
            stmt.find1("type", required=True), sctx)
        self._handle_child(node, stmt, sctx)

    def _rpc_action_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle rpc or action statement."""
        self._handle_child(RpcActionNode(), stmt, sctx)

    def _notification_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle notification statement."""
        self._handle_child(NotificationNode(), stmt, sctx)

    def _anydata_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle anydata statement."""
        self._handle_child(AnydataNode(), stmt, sctx)

    def _ascii_tree(self, indent: str, no_types: bool, val_count: bool) -> str:
        """Return the receiver's subtree as ASCII art."""
        def suffix(sn):
            return f" {{{sn.val_count}}}\n" if val_count else "\n"
        if not self.children:
            return ""
        cs = []
        for c in self.children:
            cs.extend(c._flatten())
        cs.sort(key=lambda x: x.qual_name)
        res = ""
        for c in cs[:-1]:
            res += (indent + c._tree_line(no_types) + suffix(c) +
                    c._ascii_tree(indent + "|  ", no_types, val_count))
        return (res + indent + cs[-1]._tree_line(no_types) + suffix(cs[-1]) +
                cs[-1]._ascii_tree(indent + "   ", no_types, val_count))

    def clear_val_counters(self) -> None:
        """Clear validation counters in the receiver and its subtree."""
        super().clear_val_counters()
        for c in self.children:
            c.clear_val_counters()


class GroupNode(InternalNode):
    """Anonymous group of schema nodes."""

    def _handle_child(self, node: SchemaNode, stmt: Statement,
                      sctx: SchemaContext) -> None:
        if not isinstance(self.parent, ChoiceNode) or isinstance(node, CaseNode):
            super()._handle_child(node, stmt, sctx)
        else:
            cn = CaseNode()
            cn.name = stmt.argument
            cn.ns = sctx.default_ns
            self._add_child(cn)
            cn._handle_child(node, stmt, sctx)

    def _pattern_entry(self) -> SchemaPattern:
        return super()._schema_pattern()

    def _flatten(self) -> List[SchemaNode]:
        res = []
        for c in self.children:
            res.extend(c._flatten())
        return res


class SchemaTreeNode(GroupNode):
    """Root node of a schema tree."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.annotations = {}  # type: Dict[QualName, Annotation]

    def data_parent(self) -> InternalNode:
        """Override the superclass method."""
        return self.parent

    def _annotation_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle annotation statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        dst = stmt.find1("description")
        self.annotations[(stmt.argument, sctx.default_ns)] = Annotation(
            DataType._resolve_type(stmt.find1("type", required=True), sctx),
            dst.argument if dst else None)


class DataNode(SchemaNode):
    """Abstract superclass for all data nodes."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_deny = DefaultDeny.none  # type: "DefaultDeny"

    def orphan_instance(self, rval: RawValue) -> "ObjectMember":
        """Return an isolated instance of the receiver.

        Args:
            rval: Raw value to be used for the returned instance.
        """
        val = self.from_raw(rval)
        return ObjectMember(self.iname(), {}, val, None, self, datetime.now())

    def split_instance_route(self, route: "InstanceRoute") -> Optional[Tuple[
            "InstanceRoute", "InstanceRoute"]]:
        """Split `route` into the part up to receiver and the rest.

        Args:
            route: Absolute instance route (the receiver should correspond to an
                instance node on this route).

        Returns:
            A tuple consisting of
                - the part of `route` from the root up to and including the
                  instance whose schema node is the receiver, and
                - the rest of `route`.
                ``None`` is returned if the receiver is not on the route.
        """
        sroute = []
        sn = self
        while sn:
            sroute.append(sn.iname())
            sn = sn.data_parent()
        i = 0
        while True:
            if not sroute:
                break
            inst = sroute.pop()
            if inst != route[i].iname():
                return None
            while True:         # skip up to next MemberName
                i += 1
                if i >= len(route) or isinstance(route[i], MemberName):
                    break
            if not sroute:
                return (InstanceRoute(route[:i]), InstanceRoute(route[i:]))
            if i >= len(route):
                return None

    def _validate(self, inst: "InstanceNode", scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Extend the superclass method."""
        if scope.value & ValidationScope.semantics.value:
            self._check_must(inst)        # must expressions
        super()._validate(inst, scope, ctype)

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        iname = self.iname()
        if iname in pnode.value:
            return pnode
        nm = pnode.put_member(iname, (None,))
        if not self.when or self.when.evaluate(nm):
            wd = self._default_value(nm, ctype, lazy)
            if wd.value is not None:
                return wd.up()
        return pnode

    def _check_must(self, inst: "InstanceNode") -> None:
        for m in self.must:
            if not m.expression.evaluate(inst):
                raise SemanticError(inst.json_pointer(), m.error_tag,
                                    m.error_message)

    def _pattern_entry(self) -> SchemaPattern:
        m = Member(self.iname(), self.content_type(), self.when)
        return m if self.mandatory else SchemaPattern.optional(m)

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + (
            "ro" if self.content_type() == ContentType.nonconfig else "rw")


class TerminalNode(SchemaNode):
    """Abstract superclass for terminal nodes in the schema tree."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.type = None  # type: DataType
        self._default = None  # type: Optional[Value]

    def content_type(self) -> ContentType:
        """Override superclass method."""
        if self._ctype:
            return self._ctype
        return (ContentType.config if self.parent.config else
                ContentType.nonconfig)

    def from_raw(self, rval: RawScalar, jptr: JSONPointer = "") -> ScalarValue:
        """Override the superclass method."""
        res = self.type.from_raw(rval)
        if res is None:
            raise RawTypeError(jptr, self.type.yang_type() + " value")
        return res

    def _node_digest(self) -> Dict[str, Any]:
        res = super()._node_digest()
        res["type"] = self.type._type_digest(self.config)
        df = self.default
        if df is not None:
            res["default"] = self.type.to_raw(df)
        return res

    def _validate(self, inst: "InstanceNode", scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Extend the superclass method."""
        if (scope.value & ValidationScope.syntax.value and
                inst.value not in self.type):
            raise YangTypeError(inst.json_pointer(), self.type.error_tag,
                                self.type.error_message)
        if (isinstance(self.type, LinkType) and        # referential integrity
                scope.value & ValidationScope.semantics.value and
                self.type.require_instance):
            try:
                tgt = inst._deref()
            except YangsonException:
                tgt = []
            if not tgt:
                raise SemanticError(inst.json_pointer(), "instance-required")
        super()._validate(inst, scope, ctype)

    def _default_value(self, inst: "InstanceNode", ctype: ContentType,
                       lazy: bool) -> "InstanceNode":
        inst.value = self.default
        return inst

    def _post_process(self) -> None:
        super()._post_process()
        self.type._post_process(self)

    def _is_identityref(self) -> bool:
        return isinstance(self.type, IdentityrefType)

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        di = self._default_instance(inst, ContentType.all)
        return [] if di is None else [self]

    def _ascii_tree(self, indent: str, no_types: bool, val_count: bool) -> str:
        return ""

    def _state_roots(self) -> List[SchemaNode]:
        return [] if self.content_type() == ContentType.config else [self]


class ContainerNode(DataNode, InternalNode):
    """Container node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.presence = False  # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return not self.presence and super().mandatory

    def _node_digest(self) -> Dict[str, Any]:
        res = super()._node_digest()
        res["presence"] = self.presence
        return res

    def _add_mandatory_child(self, node: SchemaNode):
        if not (self.presence or self.mandatory):
            self.parent._add_mandatory_child(self)
        super()._add_mandatory_child(node)

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        if self.presence:
            return pnode
        return super()._default_instance(pnode, ctype, lazy)

    def _default_value(self, inst: "InstanceNode", ctype: ContentType,
                       lazy: bool) -> Optional["InstanceNode"]:
        inst.value = ObjectValue()
        return inst if lazy else self._add_defaults(inst, ctype)

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        if self.presence:
            return []
        res = inst.put_member(self.iname(), ObjectValue())
        if self.when is None or self.when.evaluate(res):
            return [res]
        return []

    def _presence_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.presence = True

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return super()._tree_line() + ("!" if self.presence else "")


class SequenceNode(DataNode):
    """Abstract class for data nodes that represent a sequence."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.min_elements = 0  # type: int
        self.max_elements = None  # type: Optional[int]
        self.user_ordered = False  # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self.min_elements > 0

    def _validate(self, inst: "InstanceNode", scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Extend the superclass method."""
        if isinstance(inst, ArrayEntry):
            super()._validate(inst, scope, ctype)
        else:
            if scope.value & ValidationScope.semantics.value:
                self._check_list_props(inst)
                self._check_cardinality(inst)
            for e in inst:
                super()._validate(e, scope, ctype)

    def _check_cardinality(self, inst: "InstanceNode") -> None:
        if len(inst.value) < self.min_elements:
            raise SemanticError(inst.json_pointer(), "too-few-elements")
        if (self.max_elements is not None and
                len(inst.value) > self.max_elements):
            raise SemanticError(inst.json_pointer(), "too-many-elements")

    def _post_process(self) -> None:
        super()._post_process()
        if self.min_elements > 0:
            self.parent._add_mandatory_child(self)

    def _min_elements_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.min_elements = int(stmt.argument)

    def _max_elements_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        arg = stmt.argument
        if arg == "unbounded":
            self.max_elements = None
        else:
            self.max_elements = int(arg)

    def _ordered_by_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.user_ordered = stmt.argument == "user"

    def _tree_line(self, no_type: bool = False) -> str:
        """Extend the superclass method."""
        return super()._tree_line() + "*"

    def from_raw(self, rval: RawList, jptr: JSONPointer = "") -> ArrayValue:
        """Override the superclass method."""
        if not isinstance(rval, list):
            raise RawTypeError(jptr, "array")
        res = ArrayValue()
        i = 0
        for en in rval:
            i += 1
            res.append(self.entry_from_raw(en, f"{jptr}/{i}"))
        return res

    def entry_from_raw(self, rval: RawEntry, jptr: JSONPointer = "") -> EntryValue:
        """Transform a raw (leaf-)list entry into the cooked form.

        Args:
            rval: raw entry (scalar or object)
            jptr: JSON pointer of the entry

        Raises:
            NonexistentSchemaNode: If a member inside `rval` is not defined
                in the schema.
            RawTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        return super().from_raw(rval, jptr)


class ListNode(SequenceNode, InternalNode):
    """List node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.keys = []  # type: List[QualName]
        self._key_members = []
        self.unique = []  # type: List[List[SchemaRoute]]

    def _node_digest(self) -> Dict[str, Any]:
        res = super()._node_digest()
        res["keys"] = self._key_members
        return res

    def _check_list_props(self, inst: "InstanceNode") -> None:
        """Check uniqueness of keys and "unique" properties, if applicable."""
        if self.keys:
            self._check_keys(inst)
        for u in self.unique:
            self._check_unique(u, inst)

    def _check_keys(self, inst: "InstanceNode") -> None:
        ukeys = set()
        for i in range(len(inst.value)):
            en = inst.value[i]
            try:
                kval = tuple([en[k] for k in self._key_members])
            except KeyError as e:
                raise SchemaError(inst._entry(i).json_pointer(),
                                  "list-key-missing", e.args[0]) from None
            if kval in ukeys:
                raise SemanticError(inst.json_pointer(), "non-unique-key", repr(
                    kval[0] if len(kval) < 2 else kval))
            ukeys.add(kval)

    def _check_unique(self, unique: List[SchemaRoute],
                      inst: "InstanceNode") -> None:
        uvals = set()
        for en in inst:
            den = en.add_defaults()
            uval = tuple([den._peek_schema_route(sr) for sr in unique])
            if None not in uval:
                if uval in uvals:
                    raise SemanticError(inst.json_pointer(), "data-not-unique")
                else:
                    uvals.add(uval)

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        return pnode

    def _post_process(self) -> None:
        super()._post_process()
        for k in self.keys:
            kn = self.get_data_child(*k)
            self._key_members.append(kn.iname())
            if not kn._mandatory:
                kn._mandatory = True
                self._mandatory_children.add(kn)

    def _key_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.keys = []
        for k in stmt.argument.split():
            self.keys.append(sctx.schema_data.translate_node_id(k, sctx))

    def _unique_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        uspec = []
        for sid in stmt.argument.split():
            uspec.append(sctx.schema_data.sni2route(sid, sctx))
        self.unique.append(uspec)

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        keys = (" [" + " ".join([k[0] for k in self.keys]) + "]"
                if self.keys else "")
        return super()._tree_line() + keys

    def orphan_entry(self, rval: RawObject) -> "ArrayEntry":
        """Return an isolated entry of the receiver.

        Args:
            rval: Raw object to be used for the returned entry.
        """
        val = self.entry_from_raw(rval)
        return ArrayEntry(0, EmptyList(), EmptyList(), val, None, self,
                          val.timestamp)


class ChoiceNode(InternalNode):
    """Choice node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_case = None  # type: QualName
        self._mandatory = False  # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self._mandatory

    def _add_defaults(self, inst: "InstanceNode",
                      ctype: ContentType) -> "InstanceNode":
        if self.when and not self.when.evaluate(inst):
            return inst
        ac = self._active_case(inst.value)
        if ac:
            return ac._add_defaults(inst, ctype)
        elif self.default_case:
            n = dc = self.get_child(*self.default_case)
            while n is not self:
                if n.when and not n.when.evaluate(inst):
                    return inst
                n = n.parent
            return dc._add_defaults(inst, ctype)
        else:
            return inst

    def _active_case(self, value: ObjectValue) -> Optional["CaseNode"]:
        """Return receiver's case that's active in an instance node value."""
        for c in self.children:
            for cc in c.data_children():
                if cc.iname() in value:
                    return c

    def _pattern_entry(self) -> SchemaPattern:
        if not self.children:
            return Empty()
        prev = self.children[0]._schema_pattern()
        for c in self.children[1:]:
            prev = ChoicePattern(c._schema_pattern(), prev, self.name)
        prev.ctype = self.content_type()
        if not self.mandatory:
            prev = SchemaPattern.optional(prev)
        return ConditionalPattern(prev, self.when) if self.when else prev

    def _post_process(self) -> None:
        super()._post_process()
        if self._mandatory:
            self.parent._add_mandatory_child(self)

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        res = []
        if self.default_case is None:
            return res
        for cn in self.get_child(*self.default_case).children:
            res.extend(cn._default_nodes(inst))
        return res

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + (
            "ro" if self.content_type() == ContentType.nonconfig else "rw")

    def _handle_child(self, node: SchemaNode, stmt: Statement,
                      sctx: SchemaContext) -> None:
        if isinstance(node, CaseNode):
            super()._handle_child(node, stmt, sctx)
        else:
            cn = CaseNode()
            cn.name = stmt.argument
            cn.ns = sctx.default_ns
            self._add_child(cn)
            cn._handle_child(node, stmt, sctx)

    def _default_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.default_case = sctx.schema_data.translate_node_id(
            stmt.argument, sctx)

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return f"{self._tree_line_prefix()} ({self.iname()})" \
               f"{'' if self._mandatory else '?'}"


class CaseNode(InternalNode):
    """Case node."""

    def _pattern_entry(self) -> SchemaPattern:
        return super()._schema_pattern()

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return f"{self._tree_line_prefix()}:({self.iname()})"


class LeafNode(DataNode, TerminalNode):
    """Leaf node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._mandatory = False  # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self._mandatory

    @property
    def default(self) -> Optional[ScalarValue]:
        """Default value of the receiver, if any."""
        if self.mandatory:
            return None
        if self._default is not None:
            return self._default
        return self.type.default

    def _post_process(self) -> None:
        super()._post_process()
        if self._mandatory:
            self.parent._add_mandatory_child(self)
        elif self._default is not None:
            self._default = self.type.from_yang(self._default)

    def _tree_line(self, no_type: bool = False) -> str:
        res = super()._tree_line() + ("" if self._mandatory else "?")
        return res if no_type else f"{res} <{self.type}>"

    def _default_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self._default = stmt.argument


class LeafListNode(SequenceNode, TerminalNode):
    """Leaf-list node."""

    @property
    def default(self) -> Optional[ScalarValue]:
        """Default value of the receiver, if any."""
        if self.mandatory:
            return None
        if self._default is not None:
            return self._default
        return (None if self.type.default is None
                else ArrayValue([self.type.default]))

    def _yang_class(self) -> str:
        return "leaf-list"

    def _check_list_props(self, inst: "InstanceNode") -> None:
        if (self.content_type() == ContentType.config and
                len(set(inst.value)) < len(inst.value)):
            raise SemanticError(inst.json_pointer(), "repeated-leaf-list-value")

    def _default_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        if self._default is None:
            self._default = [stmt.argument]
        else:
            self._default.append(stmt.argument)

    def _post_process(self) -> None:
        super()._post_process()
        if self._default is not None:
            self._default = ArrayValue(
                [self.type.from_yang(v) for v in self._default])

    def _tree_line(self, no_type: bool = False) -> str:
        res = super()._tree_line()
        return res if no_type else f"{res} <{self.type}>"


class AnyContentNode(DataNode):
    """Abstract class for anydata or anyxml nodes."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._mandatory = False  # type: bool

    def content_type(self) -> ContentType:
        """Override superclass method."""
        return TerminalNode.content_type(self)

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self._mandatory

    def from_raw(self, rval: RawValue, jptr: JSONPointer = "") -> Value:
        """Override the superclass method."""
        def convert(val):
            if isinstance(val, list):
                res = ArrayValue([convert(x) for x in val])
            elif isinstance(val, dict):
                res = ObjectValue({x: convert(val[x]) for x in val})
            else:
                res = val
            return res
        return convert(rval)

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        return pnode

    def _tree_line(self, no_type: bool = False) -> str:
        return super()._tree_line() + ("" if self._mandatory else "?")

    def _ascii_tree(self, indent: str, no_types: bool, val_count: bool) -> str:
        return ""

    def _post_process(self) -> None:
        if self._mandatory:
            self.parent._add_mandatory_child(self)


class AnydataNode(AnyContentNode):
    """Anydata node."""
    pass


class AnyxmlNode(AnyContentNode):
    """Anyxml node."""
    pass


class RpcActionNode(SchemaTreeNode):
    """RPC or action node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_deny = DefaultDeny.none  # type: "DefaultDeny"
        self._ctype = ContentType.nonconfig

    def _handle_substatements(self, stmt: Statement, sctx: SchemaContext) -> None:
        self._add_child(InputNode(sctx.default_ns))
        self._add_child(OutputNode(sctx.default_ns))
        super()._handle_substatements(stmt, sctx)

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "-x"

    def _input_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle RPC or action input statement."""
        self.get_child("input")._handle_substatements(stmt, sctx)

    def _output_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle RPC or action output statement."""
        self.get_child("output")._handle_substatements(stmt, sctx)


class InputNode(SchemaTreeNode):
    """RPC or action input node."""

    def __init__(self, ns):
        """Initialize the class instance."""
        super().__init__()
        self._config = False
        self.name = "input"
        self.ns = ns

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "ro"


class OutputNode(SchemaTreeNode):
    """RPC or action output node."""

    def __init__(self, ns):
        """Initialize the class instance."""
        super().__init__()
        self._config = False
        self.name = "output"
        self.ns = ns

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "ro"


class NotificationNode(SchemaTreeNode):
    """Notification node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_deny = DefaultDeny.none  # type: "DefaultDeny"
        self._ctype = ContentType.nonconfig

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "-n"


from .xpathast import Expr, LocationPath, Step, Root        # NOQA
from .instance import (ArrayEntry, EmptyList, InstanceNode,  # NOQA
                       InstanceRoute, MemberName, ObjectMember)



import json
import pytest
from decimal import Decimal
from yangson import DataModel
from yangson.exceptions import (
    InvalidFeatureExpression, UnknownPrefix, NonexistentInstance,
    NonexistentSchemaNode, RawTypeError, SchemaError,
    XPathTypeError, InvalidXPath, NotSupported)
from yangson.instvalue import ArrayValue
from yangson.schemadata import SchemaContext, FeatureExprParser
from yangson.enumerations import ContentType
from yangson.xpathparser import XPathParser

tree = """+--rw (test:choiA)?
|  +--:(test:caseA)
|  |  +--rw test:contC
|  |  |  +--rw leafD? <typB(int16)>
|  |  |  +--rw llistA* <typA(int16)>
|  |  +--rw test:leafH? <ip-address(union)>
|  +--:(testb:leafQ)
|  |  +--rw testb:leafQ? <empty>
|  +--:(test:llistB)
|     +--rw test:llistB* <ip-address-no-zone(union)>
+--rw test:contA
|  +--rw anydA
|  +--rw anyxA?
|  +--rw (testb:choiB)
|  |  +--:(testb:contB)
|  |  |  +--rw testb:contB!
|  |  |     +--rw leafC <typA(int16)>
|  |  +--:(testb:leafI)
|  |  |  +--rw testb:leafI? <typB(int16)>
|  |  +--:(testb:leafN)
|  |     +--rw testb:leafN? <string>
|  +--rw leafA? <typA(int16)>
|  +--ro leafB <typA(int16)>
|  +--rw testb:leafR? <leafref>
|  +--rw testb:leafS? <instance-identifier>
|  +--rw testb:leafT? <identityref>
|  +--rw testb:leafV <typE(int16)>
|  +--rw listA* [leafE leafF]
|     +--rw contD
|     |  +---x acA
|     |  |  +--ro input
|     |  |  +--ro output
|     |  |     +--ro leafL <boolean>
|     |  +--rw contE!
|     |  |  +--rw leafJ? <empty>
|     |  |  +--rw leafP? <uint8>
|     |  |  +--rw leafU? <boolean>
|     |  +--rw leafG? <yang-identifier(string)>
|     +--rw leafE <hex-number(string)>
|     +--rw leafF <boolean>
|     +--rw leafW? <typE(leafref)>
+--rw test:contT
|  +--rw binary? <binary>
|  +--rw bits? <typD(bits)>
|  +--rw boolean? <boolean>
|  +--rw decimal64? <decimal64>
|  +--rw enumeration? <typC(enumeration)>
|  +--rw int16? <int16>
|  +--rw int32? <int32>
|  +--rw int64? <int64>
|  +--rw int8? <int8>
|  +--rw string? <string>
|  +--rw uint16? <uint16>
|  +--rw uint32? <uint32>
|  +--rw uint64? <uint64>
|  +--rw uint8? <uint8>
+--rw test:leafX? <port-number(uint16)>
+---n testb:noA
|  +--ro testb:leafO? <boolean>
+---x testb:rpcA
   +--ro input
   |  +--ro testb:leafK? <typA(int16)>
   +--ro output
      +--ro testb:llistC* <boolean>
"""


@pytest.fixture
def data_model():
    return DataModel.from_file("yang-modules/test/yang-library.json",
                               ["yang-modules/test", "yang-modules/ietf"])


@pytest.fixture
def instance(data_model):
    data = """
    {
        "test:llistB": ["::1", "127.0.0.1"],
        "test:leafX": 53531,
        "test:contA": {
            "leafB": 9,
            "listA": [{
                "leafE": "C0FFEE",
                "leafF": true,
                "contD": {
                    "leafG": "foo1-bar",
                    "contE": {
                        "leafJ": [null],
                        "leafP": 10
                    }
                }
            },
            {
                "leafE": "ABBA",
                "leafW": 9,
                "leafF": false
            }],
            "testb:leafS":
                "/test:contA/listA[leafE='C0FFEE'][leafF='true']/contD/contE/leafP",
            "testb:leafR": "C0FFEE",
            "testb:leafT": "test:CC-BY",
            "testb:leafV": 99,
            "anydA": {
                "foo:bar": [1, 2, 3]
            },
            "testb:leafN": "hi!"
        },
        "test:contT": {
            "bits": "dos cuatro",
            "decimal64": 4.50,
            "enumeration": "Hearts"
        }
    }
    """
    return data_model.from_raw(json.loads(data))


def test_schema_data(data_model):
    assert len(data_model.schema_data.implement) == 2
    assert data_model.module_set_id() == "b6d7e0614440c5ad8a7370fe46c777254d331983"
    tid = data_model.schema_data.last_revision("test")
    stid = data_model.schema_data.last_revision("subtest")
    tbid = data_model.schema_data.last_revision("testb")
    assert data_model.schema_data.modules[tid].statement.argument == "test"
    assert data_model.schema_data.translate_pname("t:foo", tbid) == ("foo", "test")
    assert data_model.schema_data.translate_pname("sd:foo", stid) == ("foo", "defs")
    with pytest.raises(UnknownPrefix):
        data_model.schema_data.translate_pname("d:foo", stid)
    assert FeatureExprParser("feA and not (not feA or feB)", data_model.schema_data, tid).parse()
    with pytest.raises(InvalidFeatureExpression):
        FeatureExprParser("feA andnot (not feA or feB)", data_model.schema_data, tid).parse()
    assert not data_model.schema_data.is_derived_from(("all-uses", "test"), ("all-uses", "test"))
    assert data_model.schema_data.is_derived_from(
        ("all-uses", "test"), ("licence-property", "test"))
    assert data_model.schema_data.is_derived_from(("CC-BY-SA", "testb"), ("share-alike", "test"))
    assert not data_model.schema_data.is_derived_from(
        ("CC-BY-SA", "testb"), ("derivatives", "test"))
    assert data_model.schema_data.is_derived_from(("CC-BY-SA", "testb"), ("all-uses", "test"))


def test_schema(data_model):
    ca = data_model.get_data_node("/test:contA")
    la = ca.get_child("leafA")
    lsta = data_model.get_data_node("/test:contA/listA")
    ada = ca.get_child("anydA", "test")
    axa = ca.get_child("anyxA", "test")
    cha = data_model.get_schema_node("/test:choiA")
    cc = cha.get_data_child("contC", "test")
    ld = data_model.get_data_node("/test:contC/leafD")
    lla = cc.get_child("llistA", "test")
    chb = data_model.get_schema_node("/test:contA/testb:choiB")
    cb = chb.get_data_child("contB", "testb")
    ln = chb.get_schema_descendant(data_model.schema_data.path2route(
        "/testb:leafN/leafN"))
    lc = cb.get_data_child("leafC", "testb")
    llb = data_model.get_schema_node("/test:choiA/llistB/llistB")
    lj = data_model.get_data_node("/test:contA/listA/contD/contE/leafJ")
    assert data_model.get_data_node("/test:contA/listA/contD/leafM") is None
    llc = data_model.get_schema_node("/testb:rpcA/output/llistC")
    ll = lsta.get_schema_descendant(data_model.schema_data.path2route(
        "test:contD/acA/output/leafL"))
    assert la.parent == chb.parent == ca
    assert ll.parent.name == "output"
    assert chb in ca._mandatory_children
    assert ada in ca._mandatory_children
    assert (ada.content_type() == axa.content_type() == la.content_type() ==
            ld.content_type() == lc.content_type() == lj.content_type() ==
            ln.content_type() == ContentType.config)
    assert ca.content_type() == cha.content_type() == ContentType.all
    assert ll.content_type() == ContentType.nonconfig
    assert lj.config and ca.config and cha.config and not ll.config
    assert la.ns == ld.ns
    assert lc.ns == "testb"
    assert ld.type.default == 111
    assert lla.type.default == 11
    assert la.type.parse_value("99") == 99
    assert not ca.presence and cb.presence and not cc.presence
    assert llb.min_elements == 2
    assert llb.max_elements == 3
    assert lla.min_elements == llc.min_elements == 0
    assert lla.max_elements is None
    assert llb.user_ordered and (not lla.user_ordered)
    assert lsta.get_schema_descendant(lsta.keys[1:]).name == "leafF"
    assert lsta.get_schema_descendant(lsta.unique[0][0]).name == "leafG"
    assert data_model.get_data_node("/test:contA/listA/contD/leafM") is None
    assert data_model.get_data_node("/testb:noA/leafO") is None


def test_tree(data_model):
    assert data_model.ascii_tree() == tree


def test_types(data_model):
    llb = data_model.get_data_node("/test:llistB").type
    assert "192.168.1.254" in llb
    assert "300.1.1.1" not in llb
    assert "127.0.1" not in llb
    assert llb.parse_value("1.2.3.4.5") is None
    assert "2001:db8:0:2::1" in llb
    assert "::1" in llb
    assert "2001::db8:0:2::1" not in llb
    ct = data_model.get_data_node("/test:contT")
    i8 = ct.get_child("int8", "test").type
    assert 100 in i8
    assert -101 not in i8
    i16 = ct.get_child("int16", "test").type
    assert -32768 in i16
    assert 32768 not in i16
    i32 = ct.get_child("int32", "test").type
    assert -2147483648 in i32
    assert 2147483648 not in i32
    i64 = ct.get_child("int64", "test").type
    assert -9223372036854775808 in i64
    assert 9223372036854775808 not in i64
    assert i64.from_raw("-6378") == -6378
    ui8 = ct.get_child("uint8", "test").type
    assert 150 in ui8
    assert 99 not in ui8
    ui16 = ct.get_child("uint16", "test").type
    assert 65535 in ui16
    assert -1 not in ui16
    ui32 = ct.get_child("uint32", "test").type
    assert 4294967295 in ui32
    assert -1 not in ui32
    ui64 = ct.get_child("uint64", "test").type
    assert 18446744073709551615 in ui64
    assert -1 not in ui64
    assert ui64.from_raw("6378") == 6378
    assert ui64.from_raw("-6378") == -6378
    d64 = ct.get_child("decimal64", "test").type
    pi = Decimal("3.141592653589793238")
    assert pi in d64
    assert 10 not in d64
    assert d64.from_raw("3.14159265358979323846264338327950288") == pi
    assert d64.canonical_string(Decimal("0")) == "0.0"
    st = ct.get_child("string", "test").type
    assert st.length.intervals == [[2, 4], [11], [12]]
    assert "hello world" in st
    assert "hello-world" not in st
    assert "h" not in st
    assert "9 \tx" in st
    assert "xx xabcdefg" not in st
    boo = ct.get_child("boolean", "test").type
    assert boo.parse_value("true")
    assert False in boo
    assert boo.canonical_string(True) == "true"
    assert boo.parse_value("boo") is None
    en = ct.get_child("enumeration", "test").type
    assert "Mars" not in en
    assert "Deimos" not in en
    assert en.enum["Hearts"] == 101
    bits = ct.get_child("bits", "test").type
    assert bits.as_int(bits.from_raw("dos cuatro")) == 10
    assert bits.parse_value("un dos") == ("un", "dos")
    assert bits.canonical_string(("cuatro", "dos")) == "dos cuatro"
    assert bits.canonical_string("un dos") is None
    assert "un" not in bits
    assert "tres" not in bits
    assert bits.bit["dos"] == 1
    bin = ct.get_child("binary", "test").type
    kun = "Příliš žluťoučký kůň úpěl ďábelské ódy."
    bv = bin.parse_value(
        b'UMWZw61sacWhIMW+bHXFpW91xI1rw70ga8' +
        b'WvxYggw7pwxJtsIMSPw6FiZWxza8OpIMOzZHku')
    assert bv.decode("utf-8") == kun
    assert bin.canonical_string(kun.encode("utf-8")) == (
        "UMWZw61sacWhIMW+bHXFpW91xI1rw70ga8" +
        "WvxYggw7pwxJtsIMSPw6FiZWxza8OpIMOzZHku")


def test_instance(data_model, instance):
    def axtest(expr, res):
        assert [i.json_pointer() for i in expr] == res
    hi = hash(instance)
    instd = instance.add_defaults()
    hix = hash(instance)
    hid = hash(instd)
    assert hi == hix
    assert hi != hid
    rid1 = data_model.parse_resource_id("/test:contA/listA=C0FFEE,true/contD/contE/leafP")
    iid1 = data_model.parse_instance_id("/test:contA/listA[1]/contD/contE/leafP")
    assert instance.peek(rid1) == instance.peek(iid1)
    rid2 = data_model.parse_resource_id("/test:llistB")
    assert len(instance.peek(rid2)) == 2
    conta = instance["test:contA"]
    la1 = conta["listA"][-1]
    lt = conta["testb:leafT"]
    assert la1.index == 1
    tbln = conta["testb:leafN"]
    inst1 = la1.put_member("leafE", "ABBA").top()
    inst2 = tbln.update("hello!").top()
    assert instance.value == inst1.value
    assert instance.value != inst2.value
    assert instance.timestamp < inst1.timestamp < inst2.timestamp
    assert inst1.json_pointer() == inst2.json_pointer() == "/"
    assert la1.namespace == "test"
    assert la1["leafE"].namespace == "test"
    assert la1["leafF"].value is False
    with pytest.raises(NonexistentInstance):
        la1["contD"]
    assert la1.json_pointer() == "/test:contA/listA/1"
    assert lt.value == ("CC-BY", "test")
    assert str(lt) == "test:CC-BY"
    assert tbln.namespace == "testb"
    assert tbln.json_pointer() == "/test:contA/testb:leafN"
    assert (instance._ancestors() == instance._preceding_siblings() ==
            instance._following_siblings() == [])
    axtest(instance._ancestors_or_self(), ["/"])
    axtest(la1._ancestors(False), ["/test:contA"])
    axtest(la1._ancestors_or_self(("listA", "test")), ["/test:contA/listA/1"])
    axtest(la1._preceding_siblings(), ["/test:contA/listA/0"])
    axtest(la1._following_siblings(), [])
    assert len(conta._children()) == 10
    axtest(la1._children(("leafF", "test")), ["/test:contA/listA/1/leafF"])
    assert len(instance._descendants(with_self=True)) == 32
    axtest(conta._descendants(("listA", "test")),
           ["/test:contA/listA/0", "/test:contA/listA/1"])
    axtest(tbln._ancestors_or_self(("leafN", "testb")), ["/test:contA/testb:leafN"])


def test_xpath(data_model, instance):
    def xptest(expr, back, res=True, node=instance, module="test"):
        mid = data_model.schema_data.last_revision(module)
        xpp = XPathParser(expr, SchemaContext(data_model.schema_data,
                                              module, mid))
        pex = xpp.parse()
        assert str(pex) == back
        assert pex.evaluate(node) == res
    conta = instance["test:contA"]
    lr = conta["testb:leafR"]
    with pytest.raises(InvalidXPath):
        xptest("foo()", "foo()")
    with pytest.raises(NotSupported):
        xptest("id()", "id()")
    xptest("true()", "true()")
    xptest("false()", "false()", False)
    xptest("1 div 0", "1.0 div 0.0", float('inf'))
    xptest("-1 div 0", "-1.0 div 0.0", float('-inf'))
    xptest("string(0 div 0)", "string(0.0 div 0.0)", "NaN")
    xptest("5 mod 2", "5.0 mod 2.0", 1)
    xptest("5 mod -2", "5.0 mod -2.0", 1)
    xptest("- 5 mod 2", "-5.0 mod 2.0", -1)
    xptest("- 5 mod - 2", "-5.0 mod -2.0", -1)
    xptest("count(t:llistB)", "count(test:llistB)", 2)
    xptest("count(*)", "count(*)", 10, conta)
    xptest("count(*[. > 10])", "count(*[. > 10.0])", 2, conta)
    xptest("-leafA", "-test:leafA", -11, conta)
    xptest(" - - leafA", "test:leafA", 11, conta)
    xptest("llistB = '::1'", 'test:llistB = "::1"')
    xptest("llistB != '::1'", 'test:llistB != "::1"')
    xptest("not(llistB = '::1')", 'not(test:llistB = "::1")', False)
    xptest("llistB[position() = 2]", "test:llistB[position() = 2.0]",
           "127.0.0.1")
    xptest("count(child::llistB/following-sibling::*)",
           "count(test:llistB/following-sibling::*)", 1)
    xptest("leafA > leafB", "test:leafA > test:leafB", node=conta)
    xptest("leafA mod leafB", "test:leafA mod test:leafB", 2, node=conta)
    xptest("listA/contD/contE/leafJ = ''",
           'test:listA/test:contD/test:contE/test:leafJ = ""',
           node=conta)
    xptest("""listA[leafE='C0FFEE' ][ leafF = 'true']
           /contD/contE/leafP = 10""",
           'test:listA[test:leafE = "C0FFEE"][test:leafF = "true"]'
           '/test:contD/test:contE/test:leafP = 10.0',
           node=conta)
    xptest("listA/contD/contE/leafP < leafA | leafB",
           "test:listA/test:contD/test:contE/test:leafP < test:leafA | "
           "test:leafB", node=conta)
    xptest("listA/contD/contE/leafP > leafA | leafB",
           "test:listA/test:contD/test:contE/test:leafP > test:leafA | "
           "test:leafB", node=conta)
    xptest("listA/contD/contE/leafP = leafA | /contA/leafB",
           "test:listA/test:contD/test:contE/test:leafP = test:leafA | "
           "/test:contA/test:leafB", False, conta)
    xptest("/t:contA/t:listA[t:leafE=current()]/t:contD/t:leafG = 'foo1-bar'",
           "/test:contA/test:listA[test:leafE = current()]/"
           'test:contD/test:leafG = "foo1-bar"',
           node=lr, module="testb")
    xptest("../leafN = 'hi!'", '../testb:leafN = "hi!"',
           node=lr, module="testb")
    xptest("local-name()", "local-name()", "")
    xptest("name()", "name()", "")
    xptest("local-name(t:contA)", "local-name(test:contA)", "contA")
    xptest("name(t:contA)", "name(test:contA)", "test:contA")
    xptest("local-name()", "local-name()", "leafR", lr)
    xptest("name()", "name()", "testb:leafR", lr)
    xptest("name(../t:listA)", "name(../test:listA)", "listA", lr, "testb")
    xptest("count(descendant-or-self::*)", "count(descendant-or-self::*)", 32)
    xptest("count(descendant::t:leafE)", "count(descendant::test:leafE)", 2)
    xptest("count(preceding-sibling::*)", "count(preceding-sibling::*)",
           0, lr, "testb")
    xptest("count(following-sibling::*)", "count(following-sibling::*)",
           0, lr, "testb")
    xptest("count(descendant-or-self::contA/descendant-or-self::contA)",
           "count(descendant-or-self::test:contA/"
           "descendant-or-self::test:contA)",
           1, conta)
    xptest("count(descendant-or-self::contA/descendant::contA)",
           "count(descendant-or-self::test:contA/descendant::test:contA)",
           0, conta)
    xptest("listA[last()-1]/following-sibling::*/leafE = 'ABBA'",
           'test:listA[last() - 1.0]/following-sibling::*/test:leafE = "ABBA"',
           node=conta)
    xptest("count(//contD/parent::*/following-sibling::*/*)",
           "count(//test:contD/parent::*/following-sibling::*/*)", 4)
    xptest("//leafP = 10", "//test:leafP = 10.0")
    xptest("""count(listA[leafE = 'C0FFEE' and leafF = true()]//
           leafP/ancestor::node())""",
           'count(test:listA[test:leafE = "C0FFEE" and test:leafF = true()]//'
           "test:leafP/ancestor::node())", 5, conta)
    xptest("../* > 9", "../* > 9.0", node=lr, module="testb")
    xptest("local-name(ancestor-or-self::contA)",
           "local-name(ancestor-or-self::test:contA)",
           "contA", conta)
    xptest("string(1.0)", "string(1.0)", "1")
    xptest("string(true())", "string(true())", "true")
    xptest("string(1 = 2)", "string(1.0 = 2.0)", "false")
    xptest("string(t:contT/t:decimal64)", "string(test:contT/test:decimal64)",
           "4.5")
    xptest("string()", "string()", "C0FFEE", lr)
    xptest("concat(../t:leafA, 'foo', ., true())",
           'concat(../test:leafA, "foo", ., true())',
           "11fooC0FFEEtrue", lr, "testb")
    with pytest.raises(InvalidXPath):
        xptest("concat()", "concat()")
    xptest("starts-with(., 'C0F')", 'starts-with(., "C0F")', True, lr, "testb")
    xptest('starts-with(//listA//leafP, "1")',
           'starts-with(//test:listA//test:leafP, "1")')
    xptest("contains(., '0FF')", 'contains(., "0FF")', True, lr, "testb")
    xptest("not(contains(../leafN, '!!'))",
           'not(contains(../testb:leafN, "!!"))', True, lr, "testb")
    xptest("substring-before(//decimal64, '.')",
           'substring-before(//test:decimal64, ".")', "4")
    xptest("substring-after(//decimal64, '.')",
           'substring-after(//test:decimal64, ".")', "5")
    xptest("substring('12345', 1.5, 2.6)",
           'substring("12345", 1.5, 2.6)', "234")
    xptest("substring('12345', 0, 3)", 'substring("12345", 0.0, 3.0)', "12")
    xptest("substring('12345', 0 div 0, 3)",
           'substring("12345", 0.0 div 0.0, 3.0)', "")
    xptest("substring('12345', 1.0, 0.0 div 0.0)",
           'substring("12345", 1.0, 0.0 div 0.0)', "")
    xptest("substring('12345', -42, 1 div 0)",
           'substring("12345", -42.0, 1.0 div 0.0)', "12345")
    xptest("substring('12345', -1 div 0, 1 div 0)",
           'substring("12345", -1.0 div 0.0, 1.0 div 0.0)', "")
    xptest("substring('12345', -1 div 0)", 'substring("12345", -1.0 div 0.0)',
           "12345")
    xptest('substring(//listA[last()]/leafE, "3")',
           'substring(//test:listA[last()]/test:leafE, "3")', "BA")
    xptest("string-length(llistB)", "string-length(test:llistB)", 3)
    xptest("string-length() = 6", "string-length() = 6.0", node=lr)
    xptest("""normalize-space('  \tfoo   bar
           baz    ')""",
           'normalize-space("  &#9;foo   bar&#10;           baz    ")',
           "foo bar baz")
    xptest("translate(., 'ABCDEF', 'abcdef')",
           'translate(., "ABCDEF", "abcdef")', "c0ffee", lr)
    xptest("translate('--abcd--', 'abc-', 'ABC')",
           'translate("--abcd--", "abc-", "ABC")', "ABCd")
    xptest("boolean(foo)", "boolean(test:foo)", False)
    xptest("boolean(descendant::t:leafE)", "boolean(descendant::test:leafE)")
    xptest("boolean(10 mod 2)", "boolean(10.0 mod 2.0)", False)
    xptest("boolean(string(llistB))", "boolean(string(test:llistB))")
    xptest("number(leafA)", "number(test:leafA)", 11, conta)
    xptest("string(number())", "string(number())", "NaN", lr, "testb")
    xptest("string(number('foo'))", 'string(number("foo"))', "NaN")
    xptest("number(true()) = 1", "number(true()) = 1.0")
    xptest("number(false()) = 0", "number(false()) = 0.0")
    xptest("sum(leafA | leafB)", "sum(test:leafA | test:leafB)", 20, conta)
    xptest("string(sum(//leafE))", "string(sum(//test:leafE))", "NaN")
    xptest("sum(//leafF)", "sum(//test:leafF)", 1)
    with pytest.raises(XPathTypeError):
        xptest("sum(42)", "sum(42.0)")
    xptest("floor(t:contT/t:decimal64)", "floor(test:contT/test:decimal64)", 4)
    xptest("ceiling(t:contT/t:decimal64)",
           "ceiling(test:contT/test:decimal64)", 5)
    xptest("round(t:contT/t:decimal64)", "round(test:contT/test:decimal64)", 5)
    xptest("round(- 6.5)", "round(-6.5)", -6)
    xptest("round(1 div 0)", "round(1.0 div 0.0)", float("inf"))
    xptest("round(-1 div 0)", "round(-1.0 div 0.0)", float("-inf"))
    xptest("string(round(0 div 0))", "string(round(0.0 div 0.0))", "NaN")
    xptest("re-match(//t:leafE, '[0-9a-fA-F]*')",
           're-match(//test:leafE, "[0-9a-fA-F]*")')
    xptest("re-match(count(//t:leafE), '[0-9]*')",
           're-match(count(//test:leafE), "[0-9]*")')
    xptest(r"re-match('1.22.333', '\d{1,3}\.\d{1,3}\.\d{1,3}')",
           r're-match("1.22.333", "\d{1,3}\.\d{1,3}\.\d{1,3}")')
    xptest("re-match('aaax', 'a*')", 're-match("aaax", "a*")', False)
    xptest("re-match('a\nb', '.*')", 're-match("a&#10;b", ".*")', False)
    xptest("re-match('a\nb', '[a-z\n]*')",
           're-match("a&#10;b", "[a-z&#10;]*")')
    xptest("deref(.)/../t:leafF", "deref(.)/../test:leafF", True, lr, "testb")
    xptest("deref(../leafS)", "deref(../testb:leafS)", 10, lr, "testb")
    xptest("count(deref(../leafS) | ../leafN)",
           "count(deref(../testb:leafS) | ../testb:leafN)", 2, lr, "testb")
    xptest("derived-from-or-self(../leafT, 't:CC-BY')",
           'derived-from-or-self(../testb:leafT, "t:CC-BY")',
           True, lr, "testb")
    xptest("derived-from(../leafT, 't:CC-BY')",
           'derived-from(../testb:leafT, "t:CC-BY")', False, lr, "testb")
    xptest("derived-from(../leafT, 't:derivatives')",
           'derived-from(../testb:leafT, "t:derivatives")', True, lr, "testb")
    xptest("derived-from(../leafT, 't:share-alike')",
           'derived-from(../testb:leafT, "t:share-alike")', False, lr, "testb")
    xptest("derived-from(../leafT, 't:licence-property')",
           'derived-from(../testb:leafT, "t:licence-property")',
           True, lr, "testb")
    xptest("derived-from(., 't:CC-BY')", 'derived-from(., "t:CC-BY")',
           False, lr, "testb")
    xptest("derived-from(., 'CC-BY')", 'derived-from(., "CC-BY")',
           False, conta)
    xptest("enum-value(//enumeration)", "enum-value(//test:enumeration)", 101)
    xptest("string(enum-value(foo))", "string(enum-value(test:foo))", "NaN")
    xptest("string(enum-value(.))", "string(enum-value(.))", "NaN", conta)
    xptest("string(enum-value(.))", "string(enum-value(.))",
           "NaN", lr, "testb")
    xptest("bit-is-set(//bits, 'dos') and bit-is-set(//bits, 'cuatro')",
           'bit-is-set(//test:bits, "dos") and '
           'bit-is-set(//test:bits, "cuatro")')
    xptest("not(bit-is-set(foo, bar))", "not(bit-is-set(test:foo, test:bar))")
    xptest("bit-is-set(., 'dos')", 'bit-is-set(., "dos")', False, conta)


def test_instance_paths(data_model, instance):
    rid1 = data_model.parse_resource_id("/test:contA/testb:leafN")
    rid2 = data_model.parse_resource_id(
        "/test:contA/listA=C0FFEE,true/contD/contE")
    iid1 = data_model.parse_instance_id("/test:contA/testb:leafN")
    iid2 = data_model.parse_instance_id(
        "/test:contA/listA[leafE='C0FFEE'][leafF = 'true']/contD/contE")
    iid3 = data_model.parse_instance_id("/test:contA/listA[1]/contD/contE")
    bad_pth = "/test:contA/listA=ABBA,true/contD/contE"
    assert instance.peek(rid1) == instance.peek(iid1) == "hi!"
    assert (instance.goto(rid2)["leafP"].value ==
            instance.goto(iid2)["leafP"].value ==
            instance.goto(iid3)["leafP"].value == 10)
    with pytest.raises(NonexistentSchemaNode):
        data_model.parse_resource_id("/test:contA/leafX")
    assert str(data_model.parse_instance_id(
        "/test:contA/llX[. = 'foo']")) == '/test:contA/llX[.="foo"]'
    assert instance.peek(data_model.parse_resource_id(bad_pth)) is None
    with pytest.raises(NonexistentInstance):
        instance.goto(data_model.parse_resource_id(bad_pth))


def test_edits(data_model, instance):
    laii = data_model.parse_instance_id("/test:contA/listA")
    la = instance.goto(laii)
    inst1 = la[1].update(
        {"leafE": "B00F", "leafF": False}, raw=True).top()
    assert instance.peek(laii)[1]["leafE"] == "ABBA"
    assert inst1.peek(laii)[1]["leafE"] == "B00F"
    modla = la.delete_item(1)
    assert len(modla.value) == 1
    llb1 = instance["test:llistB"][1]
    modllb = llb1.update("2001:db8:0:2::1", raw=True).up()
    assert modllb.value == ArrayValue(["::1", "2001:db8:0:2::1"])
    with pytest.raises(RawTypeError):
        llb1.update("2001::2::1", raw=True)


def test_validation(instance):
    assert instance.validate(ctype=ContentType.all) is None
    inst2 = instance.put_member("testb:leafQ", "ABBA").top()
    with pytest.raises(SchemaError):
        inst2.validate(ctype=ContentType.all)



# Copyright © 2016-2019 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Annotated YANG constraints with custom error tags and error messages.

This module implements the following classes:

* Constraint: Abstract class representing annotated YANG constraints.
* Intervals: Class representing a sequence of numeric intervals.
* Pattern: Class representing regular expression pattern.
* Must: Class representing the constraint specified by a "must" statement.
"""

import decimal
import re
from typing import Callable, List, Optional, Union
from pyxb.utils.xmlre import RegularExpressionError, XMLToPython

from .exceptions import InvalidArgument
from .xpathast import Expr

# Type aliases
Number = Union[int, decimal.Decimal]
"""Union of numeric classes appearing in interval constraints."""

Interval = List[Number]
"""Numeric interval consisting either of one number or a pair of bounds."""


class Constraint:
    """Abstract class representing annotated YANG constraints."""

    def __init__(self, error_tag: Optional[str], error_message: Optional[str]):
        """Initialize the class instance."""
        self.error_tag = error_tag
        self.error_message = error_message


class Intervals(Constraint):
    """Class representing a sequence of numeric intervals."""

    def __init__(self, intervals: List[Interval],
                 parser: Callable[[str], Optional[Number]] = None,
                 error_tag: str = None, error_message: str = None):
        """Initialize the class instance."""
        def _pint(x):                     # default parser
            try:
                return int(x)
            except ValueError:
                return None
        super().__init__(error_tag, error_message)
        self.intervals = intervals
        self.parser = parser if parser else _pint

    def __contains__(self, value: Number):
        """Return ``True`` if the receiver contains the value."""
        for r in self.intervals:
            if len(r) == 1:
                if r[0] == value:
                    return True
            elif r[0] <= value <= r[1]:
                return True
        return False

    def __str__(self) -> str:
        """Return string representation of the receiver."""
        return " | ".join([f"{r[0]!s}..{r[-1]!s}" if len(r) > 1 else str(r[0])
                           for r in self.intervals])

    def restrict_with(self, expr: str, error_tag: str = None,
                      error_message: str = None) -> None:
        """Combine the receiver with new intervals.

        Args:
            expr: "range" or "length" expression.
            error_tag: error tag of the new expression.
            error_message: error message for the new expression.

        Raises:
            InvalidArgument: If parsing of `expr` fails.
        """
        def parse(x: str) -> Number:
            res = self.parser(x)
            if res is None:
                raise InvalidArgument(expr)
            return res

        def simpl(rng: List[Number]) -> List[Number]:
            return ([rng[0]] if rng[0] == rng[1] else rng)

        def to_num(xs): return [parse(x) for x in xs]
        lo = self.intervals[0][0]
        hi = self.intervals[-1][-1]
        ran = []
        for p in [p.strip() for p in expr.split("|")]:
            r = [i.strip() for i in p.split("..")]
            if len(r) > 2:
                raise InvalidArgument(expr)
            ran.append(r)
        if ran[0][0] != "min":
            lo = parse(ran[0][0])
        if ran[-1][-1] != "max":
            hi = parse(ran[-1][-1])
        self.intervals = (
            [simpl([lo, hi])] if len(ran) == 1 else (
                [simpl([lo, parse(ran[0][-1])])] +
                [to_num(r) for r in ran[1:-1]] +
                [simpl([parse(ran[-1][0]), hi])]))
        if error_tag:
            self.error_tag = error_tag
        if error_message:
            self.error_message = error_message


class Pattern(Constraint):
    """Class representing regular expression pattern."""

    def __init__(self, pattern: str, invert_match: bool = False,
                 error_tag: str = None,
                 error_message: str = None):
        """Initialize the class instance."""
        super().__init__(error_tag,
                         error_message if error_message else f"pattern '{pattern}'")
        self.pattern = pattern
        self.invert_match = invert_match
        try:
            self.regex = re.compile(XMLToPython(pattern))
        except RegularExpressionError:
            raise InvalidArgument(pattern) from None


class Must(Constraint):
    """Class representing the constraint specified by a "must" statement."""

    def __init__(self, expression: Expr, error_tag: str = None,
                 error_message: str = None):
        """Initialize the class instance."""
        super().__init__(
            error_tag if error_tag else "must-violation", error_message)
        self.expression = expression



"""
This script reads all *.yang files in a directory that is passed as the parameter.
Each file should contain a YANG module or submodule. The script then prints YANG
library data (JSON text) for all the modules and submodules.

If a submodule is included from a main module (with revision), then the submodule
must be present in the same directory. Otherwise an error message is printed and 
exit code is 1.
"""

import json
import os
import sys

from yangson.statement import ModuleParser

ydir = sys.argv[1]
"""Name of the directory with YANG (sub)modules."""
data_kws = ["augment", "container", "leaf", "leaf-list", "list", "rpc", "notification", "identity"]
"""Keywords of statements that contribute nodes to the schema tree."""
modmap = {}
"""Dictionary for collecting module data."""
submodmap = {}
"""Dictionary for collecting submodule data."""


def module_entry(yfile):
    """Add entry for one file containing YANG module text.

    Args:
        yfile (file): File containing a YANG module or submodule.
    """
    ytxt = yfile.read()
    mp = ModuleParser(ytxt)
    mst = mp.statement()
    submod = mst.keyword == "submodule"
    import_only = True
    rev = ""
    features = []
    includes = []
    rec = {}
    for sst in mst.substatements:
        if not rev and sst.keyword == "revision":
            rev = sst.argument
        elif import_only and sst.keyword in data_kws:
            import_only = False
        elif sst.keyword == "feature":
            features.append(sst.argument)
        elif submod:
            continue
        elif sst.keyword == "namespace":
            rec["namespace"] = sst.argument
        elif sst.keyword == "include":
            rd = sst.find1("revision-date")
            includes.append((sst.argument, rd.argument if rd else None))
    rec["import-only"] = import_only
    rec["features"] = features
    if submod:
        rec["revision"] = rev
        submodmap[mst.argument] = rec
    else:
        rec["includes"] = includes
        modmap[(mst.argument, rev)] = rec


def main():
    for infile in os.listdir(ydir):
        if not infile.endswith(".yang"):
            continue
        with open(f"{ydir}/{infile}", "r", encoding="utf-8") as yf:
            module_entry(yf)
    marr = []
    for (yam, mrev) in modmap:
        men = {"name": yam, "revision": mrev}
        sarr = []
        mrec = modmap[(yam, mrev)]
        men["namespace"] = mrec["namespace"]
        fts = mrec["features"]
        imp_only = mrec["import-only"]
        for (subm, srev) in mrec["includes"]:
            sen = {"name": subm}
            try:
                srec = submodmap[subm]
            except KeyError:
                print(f"Submodule {subm} not available.", file=sys.stderr)
                return 1
            if srev is None or srev == srec["revision"]:
                sen["revision"] = srec["revision"]
            else:
                print(f"Submodule {subm} revision mismatch.", file=sys.stderr)
                return 1
            imp_only = imp_only or srec["import-only"]
            fts += srec["features"]
            sarr.append(sen)
        if fts:
            men["feature"] = fts
        if sarr:
            men["submodule"] = sarr
        men["conformance-type"] = "import" if imp_only else "implement"
        marr.append(men)
    res = {
        "ietf-yang-library:modules-state": {
            "module-set-id": "",
            "module": marr
        }
    }
    print(json.dumps(res, indent=2))
    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright © 2016, 2017 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Structured values of instance nodes.

This module implements the following classes:

* StructuredValue: Abstract class for structured values of instance nodes.
* ArrayValue: Cooked array value of an instance node.
* ObjectValue: Cooked object value of an instance node.
"""

from datetime import datetime
from typing import Dict, List, Union
from .typealiases import InstanceName, PrefName, ScalarValue

# Type aliases
Value = Union[ScalarValue, "ArrayValue", "ObjectValue"]
"""All possible types of cooked values (scalar and structured)."""

EntryValue = Union[ScalarValue, "ObjectValue"]
"""Type of the value a list ot leaf-list entry."""

InstanceKey = Union[InstanceName, int]
"""Index of an array entry or name of an object member."""

MetadataObject = Dict[PrefName, ScalarValue]
"""Metadata object [RFC 7952]_."""


class StructuredValue:
    """Abstract class for array and object values."""

    def __init__(self, ts: datetime):
        """Initialize class instance.

        Args:
        :param ts: creation timestamp
        """
        self.timestamp = ts if ts else datetime.now()

    def copy(self) -> "StructuredValue":
        """Return a shallow copy of the receiver."""
        return self.__class__(super().copy(), datetime.now())

    def __setitem__(self, key: InstanceKey, value: Value) -> None:
        super().__setitem__(key, value)
        self.timestamp = datetime.now()

    def __eq__(self, val: "StructuredValue") -> bool:
        """Return ``True`` if the receiver equal to `val`.

        Args:
        :param val: value to compare
        """
        return self.__class__ == val.__class__ and hash(self) == hash(val)

    def __hash__(self) -> int:
        """Return hash value for the receiver."""
        raise NotImplementedError()


class ArrayValue(StructuredValue, list):
    """This class represents cooked array values."""

    def __init__(self, val: List[EntryValue] = [], ts: datetime = None):
        StructuredValue.__init__(self, ts)
        list.__init__(self, val)

    def __hash__(self) -> int:
        """Return hash value for the receiver."""
        return tuple([x.__hash__() for x in self]).__hash__()


class ObjectValue(StructuredValue, dict):
    """This class represents cooked object values."""

    def __init__(self, val: Dict[InstanceName, Value] = {},
                 ts: datetime = None):
        StructuredValue.__init__(self, ts)
        dict.__init__(self, val)

    def __hash__(self) -> int:
        """Return hash value for the receiver."""
        sks = sorted(self.keys())
        return tuple([(k, self[k].__hash__()) for k in sks]).__hash__()



# Copyright © 2016-2019 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Exceptions used by the Yangson library.

This module defines the following exceptions:

* :exc:`AnnotationException`: Base class for exceptions related to metadata annotations.
* :exc:`BadSchemaNodeType`: A schema node is of a wrong type.
* :exc:`BadYangLibraryData`: Invalid YANG library data.
* :exc:`CyclicImports`: Imports of YANG modules form a cycle.
* :exc:`DefinitionNotFound`: Requested definition does not exist.
* :exc:`EndOfInput`: Unexpected end of input.
* :exc:`FeaturePrerequisiteError`: Pre-requisite feature isn't supported.
* :exc:`InstanceException`: Base class for exceptions related to operations
  on instance nodes.
* :exc:`InstanceValueError`: The instance value is incompatible with the called method.
* :exc:`InvalidArgument`: Invalid argument of a statement.
* :exc:`InvalidFeatureExpression`: Invalid if-feature expression.
* :exc:`InvalidKeyValue`: Invalid list key or leaf-list value.
* :exc:`InvalidLeafrefPath`: A leafref path is incorrect.
* :exc:`InvalidSchemaPath`: Invalid schema path
* :exc:`InvalidXPath`: An XPath expression is invalid.
* :exc:`MissingAnnotationTarget`: Instance node that is being annotated doesn't exist.
* :exc:`MissingAugmentTarget`: Target schema node for an augment is missing.
* :exc:`MissingModule`: Abstract exception class – a module is missing.
* :exc:`ModuleContentMismatch`: Abstract exception class – unexpected module name or revision.
* :exc:`ModuleNameMismatch`: The module name doesn't match the expected name.
* :exc:`ModuleNotFound`: A module not found.
* :exc:`ModuleRevisionMismatch`: The module revision doesn't match the expected revision.
* :exc:`ModuleNotImplemented`: A module is not implemented in the data model.
* :exc:`ModuleNotImported`: A module is not imported.
* :exc:`ModuleNotRegistered`: An imported module is not registered in YANG library.
* :exc:`MultipleImplementedRevisions`: A module has multiple implemented revisions.
* :exc:`NonexistentInstance`: Attempt to access an instance node that doesn't
  exist.
* :exc:`NonDataNode`: Attempt to access an instance of non-data node
  (rpc/action/notification).
* :exc:`NonexistentSchemaNode`: A schema node doesn't exist.
* :exc:`NotSupported`: A given XPath 1.0 feature isn't (currently) supported.
* :exc:`ParserException`: Base class for parser exceptions.
* :exc:`RawDataError`: Abstract exception class for errors in raw data.
* :exc:`RawMemberError`: Object member in raw data doesn't exist in the schema.
* :exc:`RawTypeError`: Raw data value is of incorrect type.
* :exc:`SchemaError`: An instance violates a schema constraint.
* :exc:`SchemaNodeException`: Abstract exception class for schema node errors.
* :exc:`SemanticError`: An instance violates a semantic rule.
* :exc:`StatementNotFound`: Required statement does not exist.
* :exc:`UndefinedAnnotation`: Undefined annotation is used.
* :exc:`UnexpectedInput`: Unexpected input.
* :exc:`UnknownPrefix`: Unknown namespace prefix.
* :exc:`ValidationError`: Abstract exception class for instance validation errors.
* :exc:`XPathTypeError`: A subexpression is of a wrong type.
* :exc:`YangsonException`: Base class for all Yangson exceptions.
* :exc:`YangTypeError`: A scalar value is of incorrect type.
"""

from .typealiases import (InstanceName, JSONPointer, ModuleId, PrefName,
                          QualName, ScalarValue, YangIdentifier)


class YangsonException(Exception):
    """Base class for all Yangson exceptions."""
    pass


class AnnotationException(YangsonException):
    """Abstract class for exceptions related to metadata annotations."""

    def __init__(self, path: JSONPointer):
        self.path = path


class MissingAnnotationTarget(AnnotationException):
    """Instance node that is being annotated doesn't exist."""

    def __init__(self, path: JSONPointer, iname: InstanceName):
        super().__init__(path)
        self.iname = iname

    def __str__(self):
        return f"[{self.path}] no instance '{self.iname}'"


class UndefinedAnnotation(AnnotationException):
    """Undefined annotation is used."""

    def __init__(self, path: JSONPointer, aname: InstanceName):
        super().__init__(path)
        self.aname = aname

    def __str__(self):
        return f"[{self.path}] Undefined annotation '{self.aname}'"


class AnnotationTypeError(AnnotationException):
    """Type of annotation is incorrect."""

    def __init__(self, path: JSONPointer, aname: InstanceName, msg: str):
        super().__init__(path)
        self.aname = aname
        self.msg = msg

    def __str__(self):
        return f"[{self.path}] value of '{self.aname}' {self.msg}"


class InvalidArgument(YangsonException):
    """The argument of a statement is invalid."""

    def __init__(self, arg: str):
        self.argument = arg

    def __str__(self):
        return self.argument


class InvalidKeyValue(YangsonException):
    """List key or leaf-list value is invalid."""

    def __init__(self, value: ScalarValue):
        self.value = value

    def __str__(self):
        return str(self.value)


class InstanceException(YangsonException):
    """Abstract class for exceptions related to operations on instance nodes."""

    def __init__(self, path: JSONPointer, message: str):
        self.path = path
        self.message = message

    def __str__(self):
        return f"[{self.path}] {self.message}"


class InstanceValueError(InstanceException):
    """The instance value is incompatible with the called method."""
    pass


class NonexistentInstance(InstanceException):
    """Attempt to access an instance node that doesn't exist."""
    pass


class NonDataNode(InstanceException):
    """Attempt to access an instance of non-data node (rpc/action/notification)."""
    pass


class ParserException(YangsonException):
    """Base class for parser exceptions."""

    def __init__(self, parser: "Parser"):
        self.parser = parser

    def __str__(self) -> str:
        """Print parser state."""
        if "\n" in self.parser.input:
            (line, col) = self.parser.line_column()
            return f"line {line}, column {col}"
        return str(self.parser)


class EndOfInput(ParserException):
    """Unexpected end of input."""
    pass


class UnexpectedInput(ParserException):
    """Unexpected input."""

    def __init__(self, parser: "Parser", expected: str = None):
        super().__init__(parser)
        self.expected = expected

    def __str__(self) -> str:
        """Add info about expected input if available."""
        ex = "" if self.expected is None else ": expected " + self.expected
        return super().__str__() + ex


class InvalidFeatureExpression(ParserException):
    """Invalid **if-feature** expression."""
    pass


class InvalidXPath(ParserException):
    """Exception to be raised for an invalid XPath expression."""
    pass


class NotSupported(ParserException):
    """Exception to be raised for unimplemented XPath features."""

    def __init__(self, parser: "Parser", feature: str):
        super().__init__(parser)
        self.feature = feature

    def __str__(self) -> str:
        return super().str() + ": " + str(self.feature)


class MissingModule(YangsonException):
    """Abstract exception class – a module is missing."""

    def __init__(self, name: YangIdentifier, rev: str = ""):
        self.name = name
        self.rev = rev

    def __str__(self) -> str:
        if self.rev:
            return self.name + "@" + self.rev
        return self.name


class ModuleContentMismatch(YangsonException):
    """Abstract exception class – unexpected module name or revision."""

    def __init__(self, found: YangIdentifier, expected: YangIdentifier):
        self.found = found
        self.expected = expected

    def __str__(self) -> str:
        return f"'{self.found}', expected '{self.expected}'"


class ModuleNameMismatch(ModuleContentMismatch):
    """Parsed module revision doesn't match the expected revision."""
    pass


class ModuleRevisionMismatch(ModuleContentMismatch):
    """Parsed module revision doesn't match the expected revision."""
    pass


class ModuleNotFound(MissingModule):
    """A module or submodule registered in YANG library is not found."""
    pass


class ModuleNotRegistered(MissingModule):
    """A module is not registered in YANG library."""
    pass


class ModuleNotImplemented(MissingModule):
    """A module is not implemented in the data model."""
    pass


class BadYangLibraryData(YangsonException):
    """Broken YANG library data."""

    def __init__(self, message: str):
        self.message = message

    def __str__(self) -> str:
        return self.message


class InvalidSchemaPath(YangsonException):
    """Invalid schema or data path."""

    def __init__(self, path: str):
        self.path = path

    def __str__(self) -> str:
        return self.path


class MissingAugmentTarget(YangsonException):
    """Target schema node for an augment is missing.

    This can also happen in the two following cases:

    * the target node depends on a feature that is not supported
    * the module in which the target node is defined is missing
      in the YANG library, or included with a conformance type of
      ``import``.
    """

    def __init__(self, sni: str):
        self.sni = sni

    def __str__(self) -> str:
        return self.sni


class UnknownPrefix(YangsonException):
    """Unknown namespace prefix."""

    def __init__(self, prefix: YangIdentifier, mid: ModuleId):
        self.prefix = prefix
        self.mid = mid

    def __str__(self) -> str:
        return f"prefix {self.prefix} is not defined in {self.mid}"


class ModuleNotImported(YangsonException):
    """Module is not imported."""

    def __init__(self, mod: YangIdentifier, mid: ModuleId):
        self.mod = mod
        self.mid = mid

    def __str__(self) -> str:
        return f"{self.mod} not imported in {self.mid}"


class FeaturePrerequisiteError(YangsonException):
    """Pre-requisite feature is not supported."""

    def __init__(self, name: YangIdentifier, ns: YangIdentifier):
        self.name = name
        self.ns = ns

    def __str__(self) -> str:
        return f"{self.ns}:{self.name}"


class MultipleImplementedRevisions(YangsonException):
    """A module has multiple implemented revisions."""

    def __init__(self, module: YangIdentifier):
        self.module = module

    def __str__(self) -> str:
        return self.module


class CyclicImports(YangsonException):
    """YANG modules are imported in a cyclic fashion."""
    pass


class SchemaNodeException(YangsonException):
    """Abstract exception class for schema node errors."""

    def __init__(self, qn: QualName):
        self.qn = qn

    def __str__(self) -> str:
        return "/" if self.qn[0] is None else f"{self.qn[1]}:{self.qn[0]}"


class NonexistentSchemaNode(SchemaNodeException):
    """A schema node doesn't exist."""

    def __init__(self, qn: QualName, name: YangIdentifier,
                 ns: YangIdentifier = None):
        super().__init__(qn)
        self.name = name
        self.ns = ns

    def __str__(self) -> str:
        prefix = "" if self.ns is None or self.ns == self.qn[1] else self.ns + ":"
        return f"{prefix}{self.name} under {super().__str__()}"


class BadSchemaNodeType(SchemaNodeException):
    """A schema node is of a wrong type."""

    def __init__(self, qn: QualName, expected: str):
        super().__init__(qn)
        self.expected = expected

    def __str__(self) -> str:
        return super().__str__() + " is not a " + self.expected


class InvalidLeafrefPath(SchemaNodeException):
    """A leafref path is incorrect."""
    pass


class RawDataError(YangsonException):
    """Abstract exception class for errors in raw data."""

    def __init__(self, path: JSONPointer):
        self.path = path

    def __str__(self) -> JSONPointer:
        return self.path


class RawMemberError(RawDataError):
    """Object member in the raw value doesn't exist in the schema."""
    pass


class RawTypeError(RawDataError):
    """Raw value is of an incorrect type."""

    def __init__(self, path: JSONPointer, expected: str):
        super().__init__(path)
        self.message = "expected " + expected

    def __str__(self):
        return f"[{self.path}] {self.message}"


class ValidationError(YangsonException):
    """Abstract exception class for instance validation errors."""

    def __init__(self, path: JSONPointer, tag: str, message: str = None):
        self.path = path
        self.tag = tag
        self.message = message

    def __str__(self) -> str:
        msg = ": " + self.message if self.message else ""
        return f"[{self.path}] {self.tag}{msg}"


class SchemaError(ValidationError):
    """An instance violates a schema constraint."""
    pass


class SemanticError(ValidationError):
    """An instance violates a semantic rule."""
    pass


class YangTypeError(ValidationError):
    """A scalar value doesn't match its expected type."""
    pass


class StatementNotFound(YangsonException):
    """Required statement does not exist."""

    def __init__(self, parent: PrefName, kw: YangIdentifier):
        self.parent = parent
        self.keyword = kw

    def __str__(self) -> str:
        """Print the statement's keyword."""
        return f"`{self.keyword}' in `{self.parent}'"


class DefinitionNotFound(YangsonException):
    """Requested definition does not exist."""

    def __init__(self, kw: YangIdentifier, name: YangIdentifier):
        self.keyword = kw
        self.name = name

    def __str__(self) -> str:
        return f"{self.keyword} {self.name}"


class XPathTypeError(YangsonException):
    """The value of an XPath (sub)expression is of a wrong type."""

    def __init__(self, value: str):
        self.value = value

    def __str__(self) -> str:
        return self.value


from .parser import Parser      # NOQA
